from ..manipulators import discover
from ..treants import Treant
from ..trees import Veg, Leaf, Tree
from ..collections import View, Bundle

__all__ = ['Treant', 'Tree', 'Leaf', 'Bundle', 'discover', 'View']



"""
The Bundle object is the primary manipulator for Treants in aggregate. They are
returned as queries to Bundles. They offer convenience methods
for dealing with many Treants at once. Views give the same kind of aggregation
conveniences for Trees and Leaves.

"""

from __future__ import absolute_import

import os
import functools
from collections import defaultdict

import multiprocessing as mp
import glob
import fnmatch

from six import string_types
from six.moves import zip

from .trees import Tree, Leaf
from .names import TREANTDIR_NAME
from .exceptions import NotATreantError
from .metadata import Tags, Categories, AggTags, AggCategories


@functools.total_ordering
class CollectionMixin(object):
    """Mixin class for collections.

    """

    def __len__(self):
        return len(self._list())

    def __iter__(self):
        return self._list().__iter__()

    def __eq__(self, other):
        try:
            return set(self) == set(other)
        except AttributeError:
            return NotImplemented

    def __lt__(self, other):
        try:
            return set(self) < set(other)
        except AttributeError:
            return NotImplemented

    def __getitem__(self, index):
        memberlist = self._list()

        # if (is a list of bools) OR (a numpy array of bool dtype)
        if ((isinstance(index, list) and
             all(isinstance(item, bool) for item in index)) or
                (hasattr(index, 'dtype') and index.dtype == 'bool')):
            # boolean indexing, either with list or np array
            out = self.__class__([memberlist[i]
                                  for i, val in enumerate(index) if val])
        # if is list or array of ints
        elif (isinstance(index, list) or
              (hasattr(index, 'dtype') and index.dtype == 'int')):
            # fancy indexing, either with list or np array
            out = self.__class__([memberlist[item] for item in index])
        elif isinstance(index, int):
            # an index gets the member at that position
            out = memberlist[index]
        else:
            raise IndexError("Cannot index {} with given values"
                             "".format(self.__class__.__name__))
        return out

    def _membertrees(self):
        return View([member for member in self if isinstance(member, Tree)])

    def leaves(self, hidden=False):
        """Return a View of the files within the member Trees.

        Parameters
        ----------
        hidden : bool
            If True, include hidden files.

        Returns
        -------
        View
            A View giving the files in the member Trees.

        """
        return View([member.leaves(hidden=hidden)
                     for member in self._membertrees()])

    def trees(self, hidden=False):
        """Return a View of the directories within the member Trees.

        Parameters
        ----------
        hidden : bool
            If True, include hidden directories.

        Returns
        -------
        View
            A View giving the directories in the member Trees.

        """
        return View([member.trees(hidden=hidden)
                     for member in self._membertrees()])

    def children(self, hidden=False):
        """Return a View of all files and directories within the member Trees.

        Parameters
        ----------
        hidden : bool
            If True, include hidden files and directories.

        Returns
        -------
        View
            A View giving the files and directories in the member Trees.

        """
        return View([member.children(hidden=hidden)
                     for member in self._membertrees()])

    def glob(self, pattern):
        """Return a View of all child Leaves and Trees of members matching
        given globbing pattern.

        Parameters
        ----------
        pattern : string
            globbing pattern to match files and directories with

        """
        return View([member.glob(pattern) for member in self
                     if isinstance(member, Tree)])

    def draw(self, depth=None, hidden=False):
        """Print an ASCII-fied visual of all member Trees.

        Parameters
        ----------
        depth : int
            Maximum directory depth to display. ``None`` indicates no limit.
        hidden : bool
            If False, do not show hidden files; hidden directories are still
            shown if they contain non-hidden files or directories.

        """
        for member in self:
            if isinstance(member, Tree):
                member.draw(depth=depth, hidden=hidden)

    def parents(self):
        """Return a View of the parent directories for each member.

        Because a View functions as an ordered set, and some members of this
        collection may share a parent, the View of parents may contain fewer
        elements than this collection.

        """
        return View([member.parent for member in self])

    @property
    def loc(self):
        """Get a View giving Tree/Leaf at `path` relative to each Tree in
        collection.

        Use with getitem syntax, e.g. ``.loc['some name']``

        Allowed inputs are:
        - A single name
        - A list or array of names

        If directory/file does not exist at the given path, then whether a Tree
        or Leaf is given is determined by the path semantics, i.e. a trailing
        separator ("/").

        """
        if not hasattr(self, "_loc"):
            self._loc = _Loc(self)

        return self._loc

    @property
    def treeloc(self):
        """Get a View giving Tree at `path` relative to each Tree in
        collection.

        Use with getitem syntax, e.g. ``.loc['some name']``

        Allowed inputs are:
        - A single name
        - A list or array of names

        If the given path resolves to an existing file for any Tree, then a
        ``ValueError`` will be raised.

        """
        if not hasattr(self, "_treeloc"):
            self._treeloc = _TreeLoc(self)

        return self._treeloc

    @property
    def leafloc(self):
        """Get a View giving Leaf at `path` relative to each Tree in
        collection.

        Use with getitem syntax, e.g. ``.loc['some name']``

        Allowed inputs are:
        - A single name
        - A list or array of names

        If the given path resolves to an existing directory for any Tree, then
        a ``ValueError`` will be raised.

        """
        if not hasattr(self, "_leafloc"):
            self._leafloc = _LeafLoc(self)

        return self._leafloc


class View(CollectionMixin):
    """An ordered set of Trees and Leaves.

    Parameters
    ----------
    vegs : Tree, Leaf, or list
        Trees and/or Leaves to be added, which may be nested lists of Trees
        and Leaves. Trees and Leaves can be given as either objects or
        paths.

    """

    def __init__(self, *vegs):
        self._state = list()
        self._add(*vegs)

    def __repr__(self):
        names = [i.name if isinstance(i, Leaf) else i.name + '/'
                 for i in self._list()]
        return "<View({})>".format(names)

    def __getitem__(self, index):
        """Get member corresponding to the given index or slice.

        A single integer will yield a single member. Lists of either will yield
        a View with members in order by the given items. Giving a basename will
        always yield a View, since more than one Tree or Leaf may have the same
        basename and so they are not guaranteed to be unique.

        A boolean index by way of a list or numpy array can also be used to
        select out members.

        """
        memberlist = self._list()
        # we can take lists of indices or names; these return a
        # View; repeats already not respected since View functions as a
        # set
        if isinstance(index, string_types):
            # a name can be used for indexing
            # always returns a View
            out = View([memberlist[i] for i, name
                        in enumerate(self.names) if name == index])
        elif isinstance(index, slice):
            # we also take slices, obviously
            out = View(*memberlist[index])
        else:
            out = super(View, self).__getitem__(index)
        return out

    def __str__(self):
        out = "<- View ->\n"

        for member in self._list():
            out += "  {}\n".format(member.relpath)

        out += "<- ---- ->"
        return out

    def __add__(self, other):
        """Addition of View with a View or Tree/Leaf View.

        """
        if isinstance(other, (Tree, Leaf, View)):
            return View(self, other)
        else:
            raise TypeError("Right operand must be a Tree, Leaf, or View.")

    def __sub__(self, other):
        """Return a View giving the Treants in `self` that are not in `other`.

        Subtracting a Tree/Leaf from a View also works.

        """
        if isinstance(other, View):
            return View(list(set(self) - set(other)))
        elif isinstance(other, (Tree, Leaf)):
            return View(list(set(self) - set([other])))
        else:
            raise TypeError("Right operand must be a Tree, Leaf, or View.")

    def __or__(self, other):
        """Return a View giving the union of Views `self` and `other`.

        """
        if isinstance(other, View):
            return View(self, other)
        else:
            raise TypeError("Operands must be Views.")

    def __and__(self, other):
        """Return a View giving the intersection of Views `self` and `other`.

        """
        if isinstance(other, View):
            return View(list(set(self) & set(other)))
        else:
            raise TypeError("Operands must be Views.")

    def __xor__(self, other):
        """Return a View giving the symmetric difference of Views `self` and
        `other`.

        """
        if isinstance(other, View):
            return View(list(set(self) ^ set(other)))
        else:
            raise TypeError("Operands must be Views.")

    def _add(self, *vegs):
        """Add any number of members to this collection.

        :Arguments:
            *vegs*
                Trees or Leaves to add; lists, tuples, or other
                Views with Trees or Leaves will also work; strings
                giving a path (existing or not) also work, since these
                are what define Trees and Leaves

        """
        from .trees import Veg, Leaf, Tree
        from .treants import Treant

        outconts = list()
        for veg in vegs:
            if veg is None:
                pass
            elif isinstance(veg, (list, tuple)):
                self._add(*veg)
            elif isinstance(veg, (View, Bundle)):
                self._add(*list(veg))
            elif isinstance(veg, Treant):
                outconts.append(Tree(veg))
            elif isinstance(veg, Veg):
                outconts.append(veg)
            elif (isinstance(veg, string_types) and
                    (os.path.isdir(veg) or veg.endswith(os.sep))):
                tre = Tree(veg)
                outconts.append(tre)
            elif isinstance(veg, string_types):
                tre = Leaf(veg)
                outconts.append(tre)
            else:
                raise TypeError("'{}' not a valid input "
                                "for View".format(veg))

        self._add_members(*outconts)

    def _add_members(self, *members):
        """Add many members at once.

        :Arguments:
            *members*
                list of Trees and Leaves

        """
        for member in members:
            self._add_member(member)

    def _add_member(self, member):
        """Add a member to the View.

        :Arguments:
            *member*
                Tree or Leaf to add

        """
        if member not in self._state:
            self._state.append(member)

    def _list(self):
        """Return a list of members.

        """
        return list(self._state)

    @property
    def names(self):
        """List the basenames for the members in this View.

        """
        return [member.name for member in self]

    @property
    def membertrees(self):
        """A View giving only members that are Trees (or subclasses).

        """
        return self._membertrees()

    @property
    def memberleaves(self):
        """A View giving only members that are Leaves (or subclasses).

        """
        return View([member for member in self if isinstance(member, Leaf)])

    @property
    def abspaths(self):
        """List of absolute paths for the members in this View.

        """
        return [member.abspath for member in self]

    @property
    def relpaths(self):
        """List of relative paths from the current working directory for the
        members in this View.

        """
        return [member.relpath for member in self]

    @property
    def exists(self):
        """List giving existence of each member as a boolean.

        """
        return [member.exists for member in self]

    def map(self, function, processes=1, **kwargs):
        """Apply a function to each member, perhaps in parallel.

        A pool of processes is created for `processes` > 1; for example,
        with 40 members and ``processes=4``, 4 processes will be created,
        each working on a single member at any given time. When each process
        completes work on a member, it grabs another, until no members remain.

        `kwargs` are passed to the given function when applied to each member

        Parameters
        ----------
        function : function
            Function to apply to each member. Must take only a single Treant
            instance as input, but may take any number of keyword arguments.
        processes : int
            How many processes to use. If 1, applies function to each member in
            member order in serial.

        Returns
        -------
        results : list
            List giving the result of the function for each member, in member
            order. If the function returns ``None`` for each member, then only
            ``None`` is returned instead of a list.
        """
        if processes > 1:
            pool = mp.Pool(processes=processes)
            results = dict()
            results = {member.abspath: pool.apply_async(
                    function, args=(member,), kwds=kwargs) for member in self}

            pool.close()
            pool.join()

            output = {key: results[key].get() for key in results}

            # sort by member order
            results = [output[abspath] for abspath in self.abspaths]
        else:
            results = [function(member, **kwargs) for member in self]

        # check if list is all ``None``: if so, we return ``None``
        if all([(i is None) for i in results]):
            results = None

        return results

    def globfilter(self, pattern):
        """Return a View of members that match by name the given globbing
        pattern.

        Parameters
        ----------
        pattern : string
            globbing pattern to match member names with

        """
        return View([self[name] for name in
                     fnmatch.filter(self.names, pattern)])

    def make(self):
        """Make the Trees and Leaves in this View if they don't already exist.

        Returns
        -------
        View
            This View.

        """
        for member in self:
            member.make()

        return self


class Bundle(CollectionMixin):
    """An ordered set of Treants.

    Parameters
    ----------
    treants : Treant, list
        Treants to be added, which may be nested lists of Treants. Treants
        can be given as either objects or paths to directories that contain
        Treant statefiles. Glob patterns are also allowed, and all found
        Treants will be added to the collection.

    """

    def __init__(self, *treants):
        self._cache = dict()
        self._state = list()

        # add metadata objects
        self._tags = AggTags(self)
        self._categories = AggCategories(self)

        self._add(*treants)

    def __repr__(self):
        return "<Bundle({})>".format(self.names)

    def __str__(self):
        out = "<- Bundle ->\n"

        for member in self._list():
            out += "  {}\n".format(member)

        out += "<- ---- ->"
        return out

    def __getitem__(self, index):
        """Get member corresponding to the given index or slice.

        A single integer will yield a single Treant. Lists of
        either will yield a Bundle with members in order by the given items.
        Giving a name will always yield a Bundle, since names are not
        guaranteed to be unique.

        A boolean index by way of a list or numpy array can also be used to
        select out members.

        """
        # we can take lists of indices or names; these return a
        # Bundle; repeats already not respected since Bundle functions as a
        # set
        memberlist = self._list()

        if isinstance(index, string_types):
            # a name can be used for indexing
            # a name always returns a Bundle
            out = Bundle([self._list()[i]
                          for i, name in enumerate(self.names)
                          if name == index])

            if not len(out):
                raise KeyError("No name matching string selection")
        elif isinstance(index, slice):
            # we also take slices, obviously
            out = Bundle(*self._list()[index])
            out._cache.update(self._cache)
        else:
            out = super(Bundle, self).__getitem__(index)
        return out

    def __add__(self, other):
        """Addition of collections with collections or treants yields Bundle.

        """
        from .treants import Treant

        if isinstance(other, (Treant, Bundle)):
            return Bundle(self, other)
        else:
            raise TypeError("Operands must be Treant-derived or Bundles.")

    def __sub__(self, other):
        """Return a Bundle giving the Treants in `a` that are not in `b`.

        Subtracting a Treant from a collection also works.

        """
        from .treants import Treant

        if isinstance(other, Bundle):
            return Bundle(list(set(self) - set(other)))
        elif isinstance(other, Treant):
            return Bundle(list(set(self) - set([other])))
        else:
            raise TypeError("Operands must be Treant-derived or Bundles.")

    def __or__(self, other):
        """Return a Bundle giving the union of Bundles `a` and `b`.

        """
        if isinstance(other, Bundle):
            return Bundle(self, other)
        else:
            raise TypeError("Operands must be Bundles.")

    def __and__(self, other):
        """Return a Bundle giving the intersection of Bundles `a` and `b`.

        """
        if isinstance(other, Bundle):
            return Bundle(list(set(self) & set(other)))
        else:
            raise TypeError("Operands must be Bundles.")

    def __xor__(self, other):
        """Return a Bundle giving the symmetric difference of Bundles
        `a` and `b`.

        """
        if isinstance(other, Bundle):
            return Bundle(list(set(self) ^ set(other)))
        else:
            raise TypeError("Operands must be Bundles.")

    def _add(self, *treants):
        """Add any number of members to this collection.

        :Arguments:
            *treants*
                Treants to be added, which may be nested lists/tuples of
                Treants, Bundles, individual Treants or paths to existing
                Treants
        """
        from .treants import Treant

        abspaths = list()
        for treant in treants:
            if treant is None:
                pass
            elif isinstance(treant, (list, tuple, View, Bundle)):
                self._add(*treant)
            elif isinstance(treant, Treant):
                abspaths.append(treant.abspath)
                self._cache[treant.abspath] = treant
            elif isinstance(treant, Tree):
                treantdir = os.path.join(treant.abspath, TREANTDIR_NAME)
                if os.path.exists(treantdir):
                    abspaths.extend(treant.abspath)
                else:
                    raise NotATreantError("Directory '{}' is "
                                          "not a Treant".format(treant))
            elif os.path.exists(treant):
                treantdir = os.path.join(treant, TREANTDIR_NAME)
                if os.path.exists(treantdir):
                    abspaths.append(os.path.abspath(treant))
                else:
                    raise NotATreantError("Directory '{}' is "
                                          "not a Treant".format(treant))
            else:
                raise TypeError("'{}' not a valid input "
                                "for Bundle".format(treant))

        self._add_members(abspaths)

    def _remove(self, *members):
        """Remove any number of members from the collection.

        :Arguments:
            *members*
                instances, indices, names, or absolute paths of the members to
                remove

        """
        from .treants import Treant

        abspaths = self._state
        remove = list()

        for member in members:
            if isinstance(member, int):
                remove.append(abspaths[member])
            elif isinstance(member, Treant):
                remove.append(member.abspath)
            elif isinstance(member, string_types):
                # try abspaths
                abspaths = fnmatch.filter(self.abspaths, member)
                paths = [m.abspath for m in self
                         if m.abspath in abspaths]
                remove.extend(paths)
                # try names
                names = fnmatch.filter(self.names, member)
                paths = [m.abspath for m in self
                         if m.name in names]
                remove.extend(paths)
            else:
                raise TypeError('Only a Treant, index, name, or absolute '
                                'path acceptable')

        self._del_members(remove)

        # remove from cache
        for abspath in remove:
            self._cache.pop(abspath, None)

    @property
    def names(self):
        """Return a list of member names.

        :Returns:
            *names*
                list giving the name of each member, in order

        """
        names = list()
        for member in self._list():
            if member:
                names.append(member.name)
            else:
                names.append(None)

        return names

    @property
    def abspaths(self):
        """Return a list of absolute member directory paths.

        :Returns:
            *abspaths*
                list giving the absolute directory path of each member, in
                order

        """
        return [member.abspath for member in self._list()]

    @property
    def relpaths(self):
        """Return a list of relative member directory paths.

        :Returns:
            *names*
                list giving the relative directory path of each member, in
                order

        """
        return [member.relpath for member in self._list()]

    def _list(self):
        """Return a list of members.

        Note: modifications of this list won't modify the members of the
        collection!

        """
        from .treants import Treant

        abspaths = self._state

        findlist = list()
        memberlist = list()

        for abspath in abspaths:
            if abspath in self._cache:
                memberlist.append(self._cache[abspath])
            elif os.path.exists(os.path.join(abspath, TREANTDIR_NAME)):
                self._cache[abspath] = Treant(abspath)
                memberlist.append(self._cache[abspath])
            else:
                raise NotATreantError("Directory '{}' is "
                                      "not a Treant.".format(abspath))

        return memberlist

    def map(self, function, processes=1, **kwargs):
        """Apply a function to each member, perhaps in parallel.

        A pool of processes is created for *processes* > 1; for example,
        with 40 members and 'processes=4', 4 processes will be created,
        each working on a single member at any given time. When each process
        completes work on a member, it grabs another, until no members remain.

        *kwargs* are passed to the given function when applied to each member

        :Arguments:
            *function*
                function to apply to each member; must take only a single
                treant instance as input, but may take any number of keyword
                arguments

        :Keywords:
            *processes*
                how many processes to use; if 1, applies function to each
                member in member order

        :Returns:
            *results*
                list giving the result of the function for each member,
                in member order; if the function returns ``None`` for each
                member, then only ``None`` is returned instead of a list
            """
        if processes > 1:
            pool = mp.Pool(processes=processes)
            results = dict()
            results = {member.abspath: pool.apply_async(
                    function, args=(member,), kwds=kwargs) for member in self}

            output = {key: results[key].get() for key in results}

            pool.close()
            pool.join()

            # sort by member order
            results = [output[abspath] for abspath in self.abspaths]
        else:
            results = [function(member, **kwargs) for member in self]

        # check if list is all ``None``: if so, we return ``None``
        if all([(i is None) for i in results]):
            results = None

        return results

    def globfilter(self, pattern):
        """Return a Bundle of members that match by name the given globbing
        pattern.

        Parameters
        ----------
        pattern : string
            globbing pattern to match member names with

        """
        return Bundle([self[name] for name in
                      fnmatch.filter(self.names, pattern)])

    def _add_members(self, abspaths):
        """Add many members at once.

        :Arguments:
            *abspaths*
                list of abspaths

        """
        for abspath in abspaths:
            self._add_member(abspath)

    def _add_member(self, abspath):
        """Add a member to the Bundle.

        :Arguments:
            *abspath*
                absolute path to directory of new member in the filesystem

        """

        if not (abspath in self._state):
            self._state.append(abspath)

    def _del_members(self, abspaths=None, all=False):
        """Remove members from the Bundle.

        :Arguments:
            *abspaths*
                An iterable of abspaths of the members to remove
            *all*
                When True, remove all members [``False``]

        """
        if all:
            self._state = list()
        else:
            for abspath in abspaths:
                try:
                    self._state.remove(abspath)
                    self._cache.pop(abspath, None)
                except ValueError:
                    pass

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, value):
        if isinstance(value, (Tags, list, set)):
            val = list(value)
            self.tags.clear()
            self.tags.add(val)
        else:
            raise TypeError("Can only set with tags, a list, or set")

    @property
    def categories(self):
        return self._categories

    @categories.setter
    def categories(self, value):
        if isinstance(value, (Categories, dict)):
            val = dict(value)
            self.categories.clear()
            self.categories.add(val)
        else:
            raise TypeError("Can only set with categories or dict")

    def get(self, *tags, **categories):
        """Filter to only Treants which match the defined tags and categories.

        If no arguments given, the full Bundle is returned. This method should
        be thought of as a filtering, with more values specified giving only
        those Treants that match.

        Parameters
        ----------
        *tags
            Tags to match.
        **categories
            Category key, value pairs to match.

        Returns
        -------
        Bundle
            All matched Treants.

        Examples
        --------
        Doing a `get` with::

        >>> b.get('this')  # doctest: +SKIP

        is equivalent to::

        >>> b.tags.filter('this')  # doctest: +SKIP

        Finally, doing::

        >>> b.get('this', length=5)  # doctest: +SKIP

        is equivalent to::

        >>> b_n = b.tags.filter('this')  # doctest: +SKIP
        >>> b_n.categories.groupby('length')[5.0]  # doctest: +SKIP

        """
        if not (tags or categories):
            # if nothing given, return an empty version of this bundle
            return self

        output = self  # initially return self
        for c, v in categories.items():
            try:
                output &= output.categories.groupby(c)[v]
            except KeyError:
                # the key wasn't present, so we return an empty Bundle
                output = Bundle()
                break  # don't need to loop any further

        # if any tags are given that don't match anything, this will yield an
        # empty Bundle
        for t in tags:
            output &= output.tags.filter(t)

        return output


class _Loc(object):
    """Path accessor for collections."""

    def __init__(self, collection):
        self._collection = collection

    def __getitem__(self, path):
        """Get Tree/Leaf at `path` relative to each Tree in collection.

        """
        return View([t[path] for t in self._collection
                     if isinstance(t, Tree)])


class _TreeLoc(_Loc):
    """Tree accessor for collections."""

    def __getitem__(self, path):
        """Get Tree at `path` relative to each Tree in collection.

        """
        return View([t.treeloc[path] for t in self._collection
                     if isinstance(t, Tree)])


class _LeafLoc(_Loc):
    """Leaf accessor for collections."""

    def __getitem__(self, path):
        """Get Leaf at `path` relative to each Tree in collection.

        """
        return View([t.leafloc[path] for t in self._collection
                     if isinstance(t, Tree)])



"""Rsync wrapper"""

import subprocess
import os


def rsync(source, dest, compress=True, backup=False, dry=False, checksum=True,
          include=None, exclude=None, overwrite=False,
          rsync_path='/usr/bin/rsync'):
    """Wrapper function for rsync. There are some minor differences with the
    standard rsync behaviour:

    - The include option will exclude every other path.
    - The exclude statements will take place after the include statements

    Parameters
    ----------
    source : str
        Source directory for the sync
    dest : str
        Dest directory for the sync
    compress : bool
        If True, use gzip compression to reduce the data transfered over
        the network
    backup : bool
        If True, pre-existing files are renamed with a "~" extension before
        they are replaced
    dry: bool
        If True, do a dry-run. Useful for debugging purposes
    checksum: bool
        Perform a checksum to determine if file content has changed.
    include: str or list
        Paths (wildcards are allowed) to be included. If this option is used,
        every other path is excluded
    exclude: str or list
        Paths to be excluded from the copy
    overwrite: bool
        If False, files in `dest` that are newer than files in `source`
        will not be overwritten
    rsync_path: str
        Path where to find the rsync executable
    """
    opts = ['-r']  # Recursive

    if not overwrite:
        opts.append('--ignore-existing')

    if compress:
        opts.append('-z')

    if checksum:
        opts.append('-c')

    if backup:
        opts.append('-b')

    if dry:
        opts.append('--dry-run')

    if include:
        opts.extend(['--include=*/'])

        if isinstance(include, list):
            opts.extend(["--include={}".format(inc) for inc in include])
        elif isinstance(include, str):
            opts.append("--include={}".format(include))

        opts.extend(['--exclude=*'])

    if exclude:
        if isinstance(exclude, list):
            opts.extend(["--exclude={}".format(exc) for exc in exclude])
        elif isinstance(exclude, str):
            opts.append("--exclude={}".format(exclude))

    source = os.path.join(source, '')  # Add trailing slash
    cmd = [rsync_path] + opts + [source, dest]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()

    if p.returncode != 0:
        raise Exception('\n'.join([
            'Syncing error: rsync returned status code {}',
            'Standard error: {}'
            'Standard output: {}']).format(
            p.returncode, err, out))

    return cmd



# -*-encoding: utf-8 -*-

import os
from functools import wraps
from pprint import pprint

import numpy as np
import numexpr as ne
from matplotlib.tri import TriAnalyzer, Triangulation, UniformTriRefiner
from scipy.spatial import Voronoi, ConvexHull
import moderngl

from moldyn.processing.strain_CPU import StrainComputeCPU
from moldyn.simulation.builder import Model
from moldyn.utils import gl_util

def cached(f, _cache=dict()):
    """

    Parameters
    ----------
    f


    Returns
    -------

    """
    fn = f.__repr__()
    _cache[f] = ne.utils.CacheDict(12)

    @wraps(f)
    def cf(*args, **kwargs):
        argscp = []
        for arg in args:
            try:
                hash(arg)
            except:
                arg = id(arg)
            argscp.append(arg)
        for arg in kwargs.items():
            try:
                hash(arg)
            except:
                arg = (arg[0], id(arg[1]))
            argscp.append(arg)
        key = tuple(argscp)
        try:
            cv = _cache[f][key]
        except KeyError:
            cv = f(*args, **kwargs)
            _cache[f][key] = cv
        return cv
    return cf


@cached
def PDF(pos, nb_samples, rcut, bin_count):
    """
    Pair Distribution Function. Returns normalized histogram of distance between atoms.

    Parameters
    ----------
    pos : np.array
        Array containing atoms position
    nb_samples : int
        Number of atoms from which to generate the histogram
    rcut : number
        Maximum distance to consider
    bin_count : int
        Number of bins of the histogram

    Returns
    -------
    bins, hist : tuple(np.array, np.array)
        `bins` being the distances, `hist` the normalized (regarding radius) histogram

    """
    bins = np.linspace(0, rcut, bin_count)
    samples = np.random.choice(range(len(pos)), nb_samples)
    hist = np.zeros(len(bins)-1)
    for s in samples:
        sample = pos[s,:]
        dists = np.array([a for a in np.sqrt(ne.evaluate("sum((pos-sample)**2,axis=1)")) if a])
        hist += np.histogram(dists, bins=bins, weights=1/dists)[0]
    return bins[:-1], hist/nb_samples


@cached
def density(model, refinement=0):
    """
    Create a Voronoi mesh and calculate the local particle density on its vertices.

    The local density is calculated as follows:
    for each vertex, compute the density of each neighbour region as
    one over the area and assign the average of
    the neighbouring density to the vertex.

    Parameters
    ----------
    model : simulation.builder.Model
        the Model object containing
    refinement : int (defaults : 0)
        number of subdivision for refining the mesh (0 == None)
    Returns
    -------
    tri : matplotlib.tri.Triangulation
        the triangulation mesh (refined if set as)
    vert_density : numpy.array
        the array containing the local denstity associated with the tri mesh


    Example
    -------
    To plot the result using matplotlib use :

    .. code-block:: python

        import matplotlib.pyplot as plt
        tri, density = data_proc.density(model)
        plt.tricontour(tri, density) # to draw contours
        plt.tricontourf(tri, density) # ot draw filled contours
        plt.show()

    Note
    ----
    As of now, the numerical results may not be quantitatively accurate
    but should qualitatively represent the density.
    """
    vor = Voronoi(model.pos)
    vert_density = np.zeros(max(vor.vertices.shape)) # density vector
    reg_num = np.zeros(max(vor.vertices.shape)) # nbr of regions per vertex --> averaging
    for point_index, reg in enumerate(vor.point_region):
        vertices = vor.regions[reg]
        if vertices:
            if -1 not in vertices:
                area = ConvexHull(vor.vertices[vertices]).area # gets the area
                vert_density[vertices] += 1 / area # makes it a density (sort-of)
                reg_num[vertices] += 1
    vert_density /= reg_num # averaging

    # getting rid of really ugly border points
    new_vert, vert_density = (vor.vertices[vor.vertices[:, 0] >= np.min(model.pos[:, 0])],
                              vert_density[vor.vertices[:, 0] >= np.min(model.pos[:, 0])])

    new_vert, vert_density = (new_vert[new_vert[:, 0] <= np.max(model.pos[:, 0])],
                              vert_density[new_vert[:, 0] <= np.max(model.pos[:, 0])])

    new_vert, vert_density = (new_vert[new_vert[:, 1] >= np.min(model.pos[:, 1])],
                              vert_density[new_vert[:, 1] >= np.min(model.pos[:, 1])])

    new_vert, vert_density = (new_vert[new_vert[:, 1] <= np.max(model.pos[:, 1])],
                              vert_density[new_vert[:, 1] <= np.max(model.pos[:, 1])])

    # for triangulation refinement
    tri2 = Triangulation(*new_vert.T)
    if refinement:
        tri2.set_mask(TriAnalyzer(tri2).get_flat_tri_mask(0.1))
        refiner = UniformTriRefiner(tri2)
        print(len(tri2.neighbors), vert_density.shape)
        tri, vert_density = refiner.refine_field(vert_density, subdiv=refinement)
    else:
        tri, vert_density = tri2, vert_density

    return tri, vert_density

def identify(func):
    @wraps(func)
    def wrap(*args, **kwargs):
        print(repr(func))
        return func(*args, **kwargs)
    return wrap


class StrainComputeGPU:
    def __init__(self, consts):
        """

        Parameters
        ----------
        consts : dict
            Dictionary containing constants used for calculations.

        """

        self.npart = consts["npart"]

        max_layout_size = 256  # Probablement optimal (en tout cas d'après essais et guides de bonnes pratiques)
        self.groups_number = int(np.ceil(self.npart / max_layout_size))
        self.layout_size = int(np.ceil(self.npart / self.groups_number))

        consts["LAYOUT_SIZE"] = self.layout_size

        self.context = moderngl.create_standalone_context(require=430)
        #print(gl_util.source(os.path.dirname(__file__) + '/strain.glsl', consts))
        self.compute_shader = self.context.compute_shader(gl_util.source(os.path.dirname(__file__)+'/strain.glsl', consts))

        self.consts = consts


        # Buffer de positions au temps t
        self._BUFFER_P_T = self.context.buffer(reserve=2 * 4 * self.npart)
        self._BUFFER_P_T.bind_to_storage_buffer(0)

        # Buffer de positions au temps t - dt
        self._BUFFER_P_DT = self.context.buffer(reserve=2 * 4 * self.npart)
        self._BUFFER_P_DT.bind_to_storage_buffer(1)

        # Buffer d'epsilon
        self._BUFFER_E = self.context.buffer(reserve=4 * 4 * self.npart)
        self._BUFFER_E.bind_to_storage_buffer(2)

        self.array_shape = (self.npart, 2, 2)

    def set_post(self, pos):
        """

        Parameters
        ----------
        pos : np.ndarray
            Array of positions.

        Returns
        -------

        """
        self._BUFFER_P_T.write(pos.astype('f4').tobytes())

    def set_posdt(self, pos):
        """
        Parameters
        ----------
        pos : np.ndarray
            Array of positions.

        Returns
        -------

        """
        self._BUFFER_P_DT.write(pos.astype('f4').tobytes())

    def compute(self):
        """
        Compute the strain.
        Returns
        -------

        """
        self.compute_shader.run(group_x=self.groups_number)

    def get_eps(self):
        """

        Returns
        -------
        np.ndarray
            Computed inter-atomic forces.
        """
        return np.frombuffer(self._BUFFER_E.read(), dtype=np.float32).reshape(self.array_shape)


@cached
def compute_strain(model0:Model, model1:Model, rcut: float):
    """
    Compute the local deformation tensor for each atom.

    It will try to use GPU but will fallback on CPU if not available

    Parameters
    ----------
    model0: simulation.builder.Model
        The model at time t
    model1: simulation.builder.Model
        The model at time t-dt
    rcut: float
    Returns
    -------
    A vector containing the 2D deformation tensor of each atom
    (in the order of model.pos).

    Note
    ----
    Due to numerical calculation imprecision the deformation tensor may not
    be quantitatively accurate (or even symmetrical).

    """
    params = model0.params.copy()
    params["RCUT"] = rcut
    try:
        strain_compute = StrainComputeGPU(params)
    except Exception as e:
        print(e)
        strain_compute = StrainComputeCPU(params)
    strain_compute.set_post(model0.pos)
    strain_compute.set_posdt(model1.pos)

    strain_compute.compute()
    eps = strain_compute.get_eps()
    del strain_compute
    return eps





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'species_params.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_species_params(object):
    def setupUi(self, species_params):
        species_params.setObjectName("species_params")
        species_params.resize(254, 187)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Ignored)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(species_params.sizePolicy().hasHeightForWidth())
        species_params.setSizePolicy(sizePolicy)
        species_params.setWindowTitle("species_params")
        self.formLayout_2 = QtWidgets.QFormLayout(species_params)
        self.formLayout_2.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.formLayout_2.setObjectName("formLayout_2")
        self.nameLabel = QtWidgets.QLabel(species_params)
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.nameLabel)
        self.presetLabel = QtWidgets.QLabel(species_params)
        self.presetLabel.setObjectName("presetLabel")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.presetLabel)
        self.presetComboBox = QtWidgets.QComboBox(species_params)
        self.presetComboBox.setObjectName("presetComboBox")
        self.presetComboBox.addItem("")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.presetComboBox)
        self.massKgLabel = QtWidgets.QLabel(species_params)
        self.massKgLabel.setObjectName("massKgLabel")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.massKgLabel)
        self.epsilonJLabel = QtWidgets.QLabel(species_params)
        self.epsilonJLabel.setObjectName("epsilonJLabel")
        self.formLayout_2.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.epsilonJLabel)
        self.sigmaMLineEdit = QtWidgets.QLineEdit(species_params)
        self.sigmaMLineEdit.setText("")
        self.sigmaMLineEdit.setObjectName("sigmaMLineEdit")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.sigmaMLineEdit)
        self.sigmaMLabel = QtWidgets.QLabel(species_params)
        self.sigmaMLabel.setObjectName("sigmaMLabel")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.sigmaMLabel)
        self.nameLineEdit = QtWidgets.QLineEdit(species_params)
        self.nameLineEdit.setObjectName("nameLineEdit")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.nameLineEdit)
        self.massKgLineEdit = QtWidgets.QLineEdit(species_params)
        self.massKgLineEdit.setText("")
        self.massKgLineEdit.setObjectName("massKgLineEdit")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.massKgLineEdit)
        self.epsilonJLineEdit = QtWidgets.QLineEdit(species_params)
        self.epsilonJLineEdit.setText("")
        self.epsilonJLineEdit.setObjectName("epsilonJLineEdit")
        self.formLayout_2.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.epsilonJLineEdit)
        self.nameLabel.setBuddy(self.nameLineEdit)
        self.presetLabel.setBuddy(self.presetComboBox)
        self.massKgLabel.setBuddy(self.massKgLineEdit)
        self.epsilonJLabel.setBuddy(self.epsilonJLineEdit)
        self.sigmaMLabel.setBuddy(self.sigmaMLineEdit)

        self.retranslateUi(species_params)
        QtCore.QMetaObject.connectSlotsByName(species_params)
        species_params.setTabOrder(self.presetComboBox, self.nameLineEdit)
        species_params.setTabOrder(self.nameLineEdit, self.massKgLineEdit)
        species_params.setTabOrder(self.massKgLineEdit, self.sigmaMLineEdit)
        species_params.setTabOrder(self.sigmaMLineEdit, self.epsilonJLineEdit)

    def retranslateUi(self, species_params):
        _translate = QtCore.QCoreApplication.translate
        self.nameLabel.setText(_translate("species_params", "Name"))
        self.presetLabel.setText(_translate("species_params", "Preset"))
        self.presetComboBox.setItemText(0, _translate("species_params", "Choose..."))
        self.massKgLabel.setText(_translate("species_params", "Mass (kg)"))
        self.epsilonJLabel.setText(_translate("species_params", "Epsilon (J)"))
        self.sigmaMLabel.setText(_translate("species_params", "Sigma (m)"))




# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# datreant

"""
datreant --- persistent, pythonic trees for heterogeneous data
==============================================================

.. SeeAlso:: :class:`datreant.treants.Treant`

"""

# Bring some often used objects into the current namespace
from .manipulators import discover
from .treants import Treant
from .trees import Veg, Leaf, Tree
from .collections import View, Bundle

__all__ = ['Treant', 'Tree', 'Leaf', 'Bundle', 'discover', 'View']
__version__ = "1.0.2"  # NOTE: keep in sync with RELEASE in setup.py



#! /usr/bin/env python
import json
import warnings
from glob import glob
from os import path

import datreant


def convert(folder):
    treants = glob(path.join(folder, 'Treant*'))
    if len(treants) == 0:
        warnings.warn("No treant found in folder: {}".format(folder))
        return
    elif len(treants) > 1:
        warnings.warn("Multiple treants found in folder: {}\n"
                      "    Skipping conversion".format(folder))
        return

    json_file = treants[0]
    with open(json_file) as fh:
        old = json.load(fh)
    datreant.Treant(folder, categories=old['categories'], tags=old['tags'])


def main():
    import argparse
    parser = argparse.ArgumentParser(
        description="Convert pre 1.0 Treant to new Treant",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "directories",
        metavar="DIRECTORY",
        nargs="+",
        help="one or more directories that are being processed")
    args = parser.parse_args()

    for dir in args.directories:
        convert(dir)


if __name__ == '__main__':
    main()



import mock
import os
import pytest

import datreant as dtr


def test_makedirs(tmpdir):
    with tmpdir.as_cwd():
        dtr.util.makedirs('this/should/exist', exist_ok=True)
        assert os.path.exists('this/should/exist')
        assert os.path.isdir('this/should/exist')


def test_makedirs_raise(tmpdir):
    with tmpdir.as_cwd():
        dtr.util.makedirs('this/should/exist')
        with pytest.raises(OSError):
            dtr.util.makedirs('this/should/exist')


def test_makedirs_exists(tmpdir):
    # try and make a dir twice
    with tmpdir.as_cwd():
        os.mkdir('this/')
        dtr.util.makedirs('this/', exist_ok=True)
        assert os.path.exists('this/')
        assert os.path.isdir('this/')


def test_makedirs_error_catch(tmpdir):
    # mock a disk full error
    # and make sure it gets propagated through properly
    with tmpdir.as_cwd():
        with mock.patch('os.makedirs') as mp:
            mp.side_effect = OSError(os.errno.ENOSPC, 'Mock - disk full')
            # check the specific error code
            # ie check we don't mangle it enroute
            with pytest.raises(OSError) as error:
                dtr.util.makedirs('this/should/fail', exist_ok=True)
                assert error.errno == os.errno.ENOSPC



from .utils import *
from .processing import *
from . import simulation


__all__ = ["processing", "simulation", "utils"]
__version = "0.0.1"

def gui():
    import sys
    from PyQt5.QtWidgets import QApplication
    from moldyn.ui.mainwindow import MoldynMainWindow

    app = QApplication(sys.argv)

    window = MoldynMainWindow()
    window.show()

    app.exec_()


"""
User-level functions for manipulating Treants.

"""
import os
import scandir
import fnmatch
from six.moves import range

from .names import TREANTDIR_NAME


def discover(dirpath='.', depth=None, treantdepth=None):
    """Find all Treants within given directory, recursively.

    Parameters
    ----------
    dirpath : string, Tree
        Directory within which to search for Treants. May also be an existing
        Tree.
    depth : int
        Maximum directory depth to tolerate while traversing in search of
        Treants. ``None`` indicates no depth limit.
    treantdepth : int
        Maximum depth of Treants to tolerate while traversing in search
        of Treants. ``None`` indicates no Treant depth limit.

    Returns
    -------
    found : Bundle
        Bundle of found Treants.

    """
    from .collections import Bundle
    from .trees import Tree

    if isinstance(dirpath, Tree):
        if not dirpath.exists:
            raise OSError("Tree doesn't exist in the filesystem")

        dirpath = dirpath.abspath

    found = list()

    startdepth = len(dirpath.split(os.sep))
    treantdirs = set()

    for root, dirs, files in scandir.walk(dirpath):
        if (TREANTDIR_NAME in dirs):
            treantdirs.add(root)
            found.append(root)

        # depth check; if too deep, empty dirs to avoid downward traversal
        if depth is not None and len(root.split(os.sep)) - startdepth >= depth:
            for i in range(len(dirs)):
                dirs.pop()
            continue

        # Treant depth checking
        if treantdepth is not None:

            # remove Treant dirs from our set of them if we've backed out
            for treantdir in list(treantdirs):
                if treantdir not in root:
                    treantdirs.remove(treantdir)

            # actual depth check; if too deep, empty dirs to avoid downward
            # traversal
            if len(treantdirs) > treantdepth:
                for i in range(len(dirs)):
                    dirs.pop()
                continue

    return Bundle(found)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'define_external_forces.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ExternalForces(object):
    def setupUi(self, ExternalForces):
        ExternalForces.setObjectName("ExternalForces")
        ExternalForces.resize(393, 288)
        self.verticalLayout = QtWidgets.QVBoxLayout(ExternalForces)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.upGroupBox = QtWidgets.QGroupBox(ExternalForces)
        self.upGroupBox.setObjectName("upGroupBox")
        self.formLayout = QtWidgets.QFormLayout(self.upGroupBox)
        self.formLayout.setObjectName("formLayout")
        self.fractionOfSpaceLabel = QtWidgets.QLabel(self.upGroupBox)
        self.fractionOfSpaceLabel.setObjectName("fractionOfSpaceLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.fractionOfSpaceLabel)
        self.upFractionOfSpaceDoubleSpinBox = QtWidgets.QDoubleSpinBox(self.upGroupBox)
        self.upFractionOfSpaceDoubleSpinBox.setDecimals(4)
        self.upFractionOfSpaceDoubleSpinBox.setMaximum(1.0)
        self.upFractionOfSpaceDoubleSpinBox.setSingleStep(0.01)
        self.upFractionOfSpaceDoubleSpinBox.setProperty("value", 0.05)
        self.upFractionOfSpaceDoubleSpinBox.setObjectName("upFractionOfSpaceDoubleSpinBox")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.upFractionOfSpaceDoubleSpinBox)
        self.lowerLimitMLabel = QtWidgets.QLabel(self.upGroupBox)
        self.lowerLimitMLabel.setObjectName("lowerLimitMLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lowerLimitMLabel)
        self.lowerLimitMLineEdit = QtWidgets.QLineEdit(self.upGroupBox)
        self.lowerLimitMLineEdit.setObjectName("lowerLimitMLineEdit")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lowerLimitMLineEdit)
        self.xUpBtn = QtWidgets.QToolButton(self.upGroupBox)
        self.xUpBtn.setObjectName("xUpBtn")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.xUpBtn)
        self.applyXUp = QtWidgets.QCheckBox(self.upGroupBox)
        self.applyXUp.setObjectName("applyXUp")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.applyXUp)
        self.yUpBtn = QtWidgets.QToolButton(self.upGroupBox)
        self.yUpBtn.setObjectName("yUpBtn")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.yUpBtn)
        self.applyYUp = QtWidgets.QCheckBox(self.upGroupBox)
        self.applyYUp.setObjectName("applyYUp")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.applyYUp)
        self.verticalLayout.addWidget(self.upGroupBox)
        self.lowGroupBox = QtWidgets.QGroupBox(ExternalForces)
        self.lowGroupBox.setObjectName("lowGroupBox")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.lowGroupBox)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.formLayout_2 = QtWidgets.QFormLayout()
        self.formLayout_2.setObjectName("formLayout_2")
        self.fractionOfSpaceLabel_2 = QtWidgets.QLabel(self.lowGroupBox)
        self.fractionOfSpaceLabel_2.setObjectName("fractionOfSpaceLabel_2")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.fractionOfSpaceLabel_2)
        self.lowFractionOfSpaceDoubleSpinBox = QtWidgets.QDoubleSpinBox(self.lowGroupBox)
        self.lowFractionOfSpaceDoubleSpinBox.setDecimals(4)
        self.lowFractionOfSpaceDoubleSpinBox.setMaximum(1.0)
        self.lowFractionOfSpaceDoubleSpinBox.setSingleStep(0.01)
        self.lowFractionOfSpaceDoubleSpinBox.setProperty("value", 0.05)
        self.lowFractionOfSpaceDoubleSpinBox.setObjectName("lowFractionOfSpaceDoubleSpinBox")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lowFractionOfSpaceDoubleSpinBox)
        self.upperLimitMLabel = QtWidgets.QLabel(self.lowGroupBox)
        self.upperLimitMLabel.setObjectName("upperLimitMLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.upperLimitMLabel)
        self.upperLimitMLineEdit = QtWidgets.QLineEdit(self.lowGroupBox)
        self.upperLimitMLineEdit.setObjectName("upperLimitMLineEdit")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.upperLimitMLineEdit)
        self.blockLabel = QtWidgets.QLabel(self.lowGroupBox)
        self.blockLabel.setObjectName("blockLabel")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.blockLabel)
        self.blockCheckBox = QtWidgets.QCheckBox(self.lowGroupBox)
        self.blockCheckBox.setObjectName("blockCheckBox")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.blockCheckBox)
        self.verticalLayout_3.addLayout(self.formLayout_2)
        self.verticalLayout.addWidget(self.lowGroupBox)

        self.retranslateUi(ExternalForces)
        QtCore.QMetaObject.connectSlotsByName(ExternalForces)

    def retranslateUi(self, ExternalForces):
        _translate = QtCore.QCoreApplication.translate
        self.upGroupBox.setTitle(_translate("ExternalForces", "Upper zone (y+)"))
        self.fractionOfSpaceLabel.setText(_translate("ExternalForces", "Fraction of space"))
        self.lowerLimitMLabel.setText(_translate("ExternalForces", "Lower limit (m)"))
        self.xUpBtn.setText(_translate("ExternalForces", "force x component"))
        self.applyXUp.setText(_translate("ExternalForces", "Apply"))
        self.yUpBtn.setText(_translate("ExternalForces", "force y component"))
        self.applyYUp.setText(_translate("ExternalForces", "Apply"))
        self.lowGroupBox.setTitle(_translate("ExternalForces", "Lower zone (y-)"))
        self.fractionOfSpaceLabel_2.setText(_translate("ExternalForces", "Fraction of space"))
        self.upperLimitMLabel.setText(_translate("ExternalForces", "Upper limit (m)"))
        self.blockLabel.setText(_translate("ExternalForces", "Block"))




# -*-encoding: utf-8 -*-
"""

"""
"""
installer icc-rt et tbb sur les machines à processeur intel
"""

import numpy as np
import numba
import threading
import multiprocessing as mp
import ctypes


@numba.njit(nogil=True)
def force(dist, epsilon, p):
    return (-4.0 * epsilon * (6.0 * p - 12.0 * p * p)) / (dist * dist)


@numba.njit(nogil=True)
def energy(dist, epsilon, p):
    return epsilon * (4.0 * (p * p - p) + 127.0 / 4096.0)


@numba.njit(nogil=True, cache=True)
def _iterate(current_pos, i, pos, a, b, epsilon, sigma, rcut, X_PERIODIC, Y_PERIODIC, SHIFT_X, SHIFT_Y, LENGTH_X, LENGTH_Y):
    f = np.zeros((2,))
    e = 0.0
    m = 0.0
    for j in range(a, b):
        if i==j:
            continue
        distxy = current_pos - pos[j, :]

        if X_PERIODIC:
            if distxy[0] < (-SHIFT_X):
                distxy[0] += LENGTH_X
            if distxy[0] > SHIFT_X:
                distxy[0] -= LENGTH_X

        if Y_PERIODIC:
            if distxy[1] < (-SHIFT_Y):
                distxy[1] += LENGTH_Y
            if distxy[1] > SHIFT_Y:
                distxy[1] -= LENGTH_Y

        if np.abs(distxy[0])<rcut and np.abs(distxy[1])<rcut:
            dist = np.sqrt(np.sum(distxy ** 2))

            if dist < rcut:
                p = (sigma / dist) ** 6
                f += force(dist, epsilon, p)*distxy
                e += energy(dist, epsilon, p)
                m += 1.0
    return np.array((f[0], f[1], e, m))


def _par_iterate(current_pos, i, EPSILON_A, EPSILON_B, EPSILON_AB, SIGMA_A, SIGMA_B, SIGMA_AB, RCUT_A, RCUT_B,
                 RCUT_AB, N_A, NPART, LENGTH_X, LENGTH_Y, X_PERIODIC, Y_PERIODIC, SHIFT_X, SHIFT_Y):
    ret = np.zeros((4,))
    pos = np.array(_pos_array).reshape(NPART,2)
    if i < N_A:
        ret += _iterate(current_pos, i, pos, 0, N_A, EPSILON_A, SIGMA_A, RCUT_A, X_PERIODIC,
                           Y_PERIODIC, SHIFT_X, SHIFT_Y, LENGTH_X, LENGTH_Y)
        ret += _iterate(current_pos, i, pos, N_A, NPART, EPSILON_AB, SIGMA_AB, RCUT_AB,
                           X_PERIODIC, Y_PERIODIC, SHIFT_X, SHIFT_Y, LENGTH_X, LENGTH_Y)
    else:
        ret += _iterate(current_pos, i, pos, 0, N_A, EPSILON_AB, SIGMA_AB, RCUT_AB, X_PERIODIC,
                           Y_PERIODIC, SHIFT_X, SHIFT_Y, LENGTH_X, LENGTH_Y)
        ret += _iterate(current_pos, i, pos, N_A, NPART, EPSILON_B, SIGMA_B, RCUT_B, X_PERIODIC,
                           Y_PERIODIC, SHIFT_X, SHIFT_Y, LENGTH_X, LENGTH_Y)
    return ret


def _p_compute_forces(
        pool,
        _array,
        F,
        PE,
        counts,
        EPSILON_A,
        EPSILON_B,
        EPSILON_AB,
        SIGMA_A,
        SIGMA_B,
        SIGMA_AB,
        RCUT_A,
        RCUT_B,
        RCUT_AB,
        N_A,
        NPART,
        LENGTH_X,
        LENGTH_Y,
        X_PERIODIC,
        Y_PERIODIC,
        SHIFT_X,
        SHIFT_Y,
        offset,
        end):
    pos = np.array(_array).reshape(NPART,2)

    gen = ((pos[i,:], i, EPSILON_A,EPSILON_B,EPSILON_AB,SIGMA_A,SIGMA_B,SIGMA_AB,RCUT_A,RCUT_B,
            RCUT_AB,N_A,NPART,LENGTH_X,LENGTH_Y,X_PERIODIC,Y_PERIODIC, SHIFT_X, SHIFT_Y) for i in range(offset, end))

    sortie = np.array(pool.starmap(_par_iterate, gen))

    F[:,:] = sortie[:,:2]
    PE[:] = sortie[:,2]
    counts[:] = sortie[:,3]

_pos_array = None

def initProcess(array):
    global _pos_array
    _pos_array = array

class ForcesComputeCPU:
    """
    Compute module.
    Runs on CPU.
    Uses `numba` for JIT compilation and `multiprocessing` for multicore.
    See `ForcesComputeGPU` for documentation.
    """

    def __init__(self, consts, compute_npart=None, compute_offset=0):

        self.consts = consts

        self.compute_offset = max(0, compute_offset)

        self.npart = consts["NPART"]
        self.compute_npart = compute_npart or (consts["NPART"] - self.compute_offset)

        self.compute_npart = min(self.compute_npart, self.npart)

        self.array_shape = (self.npart, 2)
        self._F = np.zeros(self.array_shape, dtype=np.float32)
        self._PE = np.zeros((self.npart,), dtype=np.float32)
        self._COUNT = np.zeros((self.npart,), dtype=np.float32)

        self._thr_run = False

        self._POS = mp.Array(ctypes.c_float, self.npart * 2, lock=False)
        self._pool = mp.Pool(mp.cpu_count(), initializer=initProcess, initargs=(self._POS,))

    def __del__(self):
        self._pool.close() # pour ne pas garder des processus ouverts pour rien

    def _compute_forces(self):
        EPSILON_A = self.consts["EPSILON_A"]
        EPSILON_B = self.consts["EPSILON_B"]
        EPSILON_AB = self.consts["EPSILON_AB"]
        SIGMA_A = self.consts["SIGMA_A"]
        SIGMA_B = self.consts["SIGMA_B"]
        SIGMA_AB = self.consts["SIGMA_AB"]
        RCUT_A = self.consts["RCUT_A"]
        RCUT_B = self.consts["RCUT_B"]
        RCUT_AB = self.consts["RCUT_AB"]
        N_A = self.consts["N_A"]
        NPART = self.consts["NPART"]
        LENGTH_X = self.consts["LENGTH_X"]
        LENGTH_Y = self.consts["LENGTH_Y"]
        X_PERIODIC = self.consts["X_PERIODIC"]
        Y_PERIODIC = self.consts["Y_PERIODIC"]

        SHIFT_X = LENGTH_X / 2
        SHIFT_Y = LENGTH_Y / 2
        end = NPART
        _p_compute_forces(
            self._pool,
            self._POS,
            self._F,
            self._PE,
            self._COUNT,
            EPSILON_A,
            EPSILON_B,
            EPSILON_AB,
            SIGMA_A,
            SIGMA_B,
            SIGMA_AB,
            RCUT_A,
            RCUT_B,
            RCUT_AB,
            N_A,
            NPART,
            LENGTH_X,
            LENGTH_Y,
            X_PERIODIC,
            Y_PERIODIC,
            SHIFT_X,
            SHIFT_Y,
            self.compute_offset,
            end)

    def _join_thr(self):
        if self._thr_run:
            self._thread.join()
            self._thr_run = False

    def set_pos(self, pos):
        self._POS[:] = pos.reshape((self.npart * 2,))
        self._thr_run = True
        self._thread = threading.Thread(target=self._compute_forces)
        self._thread.start()

    def get_F(self):
        self._join_thr()
        return self._F[:,:]

    def get_PE(self):
        self._join_thr()
        return self._PE[:]

    def get_COUNT(self):
        self._join_thr()
        return self._COUNT[:]


"""Trees and Leaves: filesystem manipulation interfaces for directories and
files.

"""
from __future__ import absolute_import

import os
from functools import reduce, total_ordering
from six import string_types

from collections import OrderedDict
import scandir
from pathlib2 import Path
from asciitree import LeftAligned

from .util import makedirs
from .manipulators import discover
from .rsync import rsync


@total_ordering
class Veg(object):

    def __init__(self, filepath):
        self._path = Path(os.path.abspath(filepath))

    def __fspath__(self):
        return str(self._path.absolute())

    def __str__(self):
        return str(self.path)

    def __eq__(self, other):
        try:
            return self.abspath == other.abspath
        except AttributeError:
            return NotImplemented

    def __lt__(self, other):
        try:
            return self.abspath < other.abspath
        except AttributeError:
            return NotImplemented

    def __hash__(self):
        return hash(self.abspath)

    @property
    def exists(self):
        """Check existence of this path in filesystem.

        """
        return self.path.exists()

    @property
    def path(self):
        """Filesystem path as a :class:`pathlib2.Path`.

        """
        return self._path

    @property
    def abspath(self):
        """Absolute path.

        """
        return str(self.path.absolute())

    @property
    def relpath(self):
        """Relative path from current working directory.

        """
        return os.path.relpath(str(self.path))

    @property
    def parent(self):
        """Parent directory for this path.

        """
        return Tree(str(self.path.parent))

    @property
    def name(self):
        """Basename for this path.

        """
        return os.path.basename(os.path.abspath(self.abspath))


class Leaf(Veg):
    """A file in the filesystem."""

    def __init__(self, filepath):
        if os.path.isdir(filepath):
            raise ValueError("'{}' is an existing directory; "
                             "a Leaf must be a file".format(filepath))

        self._path = Path(os.path.abspath(filepath))

    def __repr__(self):
        return "<Leaf: '{}'>".format(self.relpath)

    def makedirs(self):
        """Make all directories along path that do not currently exist.

        Returns
        -------
        leaf : Leaf
            this leaf

        """
        makedirs(os.path.dirname(str(self.path)), exist_ok=True)

        return self

    def touch(self):
        """Make file if it doesn't exist.

        """
        self.makedirs()
        self.path.touch()

    def make(self):
        """Make the file if it doesn't exit. Equivalent to :meth:`touch`.

        Returns
        -------
        leaf : Leaf
            this leaf

        """
        self.touch()
        return self

    def read(self, size=None):
        """Read file, or up to `size` in bytes.

        Parameters
        ----------
        size : int
            extent of the file to read, in bytes

        """
        with open(self.abspath, 'r') as f:
            if size:
                out = f.read(size)
            else:
                out = f.read()
        return out


class Tree(Veg):
    """A directory."""
    def __init__(self, dirpath):
        if isinstance(dirpath, Tree):
            dirpath = dirpath.abspath

        elif os.path.isfile(dirpath):
            raise ValueError("'{}' is an existing file; "
                             "a Tree must be a directory".format(dirpath))

        self._path = Path(os.path.abspath(dirpath))

    def __repr__(self):
        return "<Tree: '{}'>".format(self.relpath)

    def __contains__(self, item):
        """Returns True if given Tree, Leaf, or plain path resolves as being
        within this Tree.

        The given path need not exist in the filesystem.

        """
        if isinstance(item, (Tree, Leaf)):
            return str(self) in str(item)
        elif isinstance(item, string_types):
            return str(self) in os.path.abspath(item)
        else:
            raise TypeError("Item must be a Tree, Leaf, or plain path")

    def __getitem__(self, path):
        """Get trees or leaves in this tree.

        Parameters
        ----------
        path : str
            Path to desired tree or leaf, relative to this tree; a list of
            paths will give a View of the resulting Trees and Leaves

        Returns
        -------
        result : Tree or Leaf, View
            Tree or Leaf corresponding to given path; if multiple paths
            given in a list, a View of the resulting Trees or Leaves is
            returned

        """
        from .collections import View

        def filt(path):
            fullpath = os.path.abspath(os.path.join(self.abspath, path))

            if (os.path.isdir(fullpath) or path.endswith(os.sep) or
                    (fullpath in self.abspath)):
                return Tree(fullpath)
            else:
                return Leaf(fullpath)

        if isinstance(path, list):
            outview = []
            for item in path:
                outview.append(filt(item))

            return View(outview)
        elif isinstance(path, string_types):
            return filt(path)
        else:
            raise ValueError('Must use a path or a list of paths')

    @property
    def loc(self):
        """Get Tree/Leaf at relative `path`.

        Use with getitem syntax, e.g. ``.loc['some name']``

        Allowed inputs are:
        - A single name
        - A list or array of names

        If directory/file does not exist at the given path, then whether a Tree
        or Leaf is given is determined by the path semantics, i.e. a trailing
        separator ("/").

        Using e.g. ``Tree.loc['some name']`` is equivalent to doing
        ``Tree['some name']``. ``.loc`` is included for parity with ``View``
        and ``Bundle`` API semantics.

        """
        if not hasattr(self, "_loc"):
            self._loc = _Loc(self)

        return self._loc

    @property
    def treeloc(self):
        """Get Tree at relative `path`.

        Use with getitem syntax, e.g. ``.treeloc['some name']``

        Allowed inputs are:
        - A single name
        - A list or array of names

        If the given path resolves to an existing file, then a ``ValueError``
        will be raised.

        """
        if not hasattr(self, "_treeloc"):
            self._treeloc = _TreeLoc(self)

        return self._treeloc

    @property
    def leafloc(self):
        """Get Leaf at relative `path`.

        Use with getitem syntax, e.g. ``.treeloc['some name']``

        Allowed inputs are:
        - A single name
        - A list or array of names

        If the given path resolves to an existing directory, then a
        ``ValueError`` will be raised.

        """
        if not hasattr(self, "_leafloc"):
            self._leafloc = _LeafLoc(self)

        return self._leafloc

    @property
    def abspath(self):
        """Absolute path of ``self.path``."""
        return str(self.path.absolute()) + os.sep

    @property
    def relpath(self):
        """Relative path of ``self.path`` from current working directory."""
        return os.path.relpath(str(self.path)) + os.sep

    @property
    def parent(self):
        """Parent directory for this path."""
        return Tree(str(self.path.parent))

    def leaves(self, hidden=False):
        """Return a View of the files in this Tree.

        Parameters
        ----------
        hidden : bool
            If True, include hidden files.

        Returns
        -------
        View
            A View with files in this Tree as members.

        """
        from .collections import View

        if not self.exists:
            raise OSError("Tree doesn't exist in the filesystem")

        for root, dirs, files in scandir.walk(self.abspath):
            if hidden:
                out = [Leaf(os.path.join(root, f)) for f in files]
            else:
                out = [Leaf(os.path.join(root, f)) for f in files
                       if f[0] != os.extsep]
            break

        out.sort()
        return View(out)

    def trees(self, hidden=False):
        """Return a View of the directories in this Tree.

        Parameters
        ----------
        hidden : bool
            If True, include hidden directories.

        Returns
        -------
        View
            A View with directories in this Tree as members.

        """
        from .collections import View

        if not self.exists:
            raise OSError("Tree doesn't exist in the filesystem")

        for root, dirs, files in scandir.walk(self.abspath):
            if hidden:
                out = [Tree(os.path.join(root, d))
                       for d in dirs]
            else:
                out = [Tree(os.path.join(root, d))
                       for d in dirs if d[0] != os.extsep]
            break

        out.sort()
        return View(out)

    def children(self, hidden=False):
        """Return a View of all files and directories in this Tree.

        Parameters
        ----------
        hidden : bool
            If True, include hidden files and directories.

        Returns
        -------
        View
            A View with files and directories in this Tree as members.

        """
        from .collections import View
        return View((self.trees(hidden=hidden) +
                     self.leaves(hidden=hidden)))

    def glob(self, pattern):
        """Return a View of all child Leaves and Trees matching given globbing
        pattern.

        Parameters
        ----------
        pattern
            globbing pattern to match files and directories with

        """
        from .collections import View

        if not self.exists:
            raise OSError("Tree doesn't exist in the filesystem")

        out = []
        for item in self.path.glob(pattern):
            out.append(self[str(item)])

        out.sort()

        return View(out)

    def walk(self, topdown=True, onerror=None, followlinks=False):
        """Walk through the contents of the tree.

        For each directory in the tree (including the root itself), yields a
        3-tuple (dirpath, dirnames, filenames).

        Parameters
        ----------
        topdown : Boolean, optional
            If False, walks directories from the bottom-up.
        onerror : function, optional
            Optional function to be called on error.
        followlinks : Boolean, optional
            If False, excludes symbolic file links.

        Returns
        -------
        generator
            Wrapped `scandir.walk()` generator yielding `datreant` objects

        """
        from .collections import View

        if not self.exists:
            raise OSError("Tree doesn't exist in the filesystem")

        for root, dirs, files in os.walk(self.abspath, topdown=topdown,
                                         onerror=onerror,
                                         followlinks=followlinks):

            # wrap results in datreant objects
            r_tree = Tree(root)
            trees = r_tree[dirs]
            leaves = r_tree[files]

            yield r_tree, trees, leaves

    def draw(self, depth=None, hidden=False):
        """Print an ASCII-fied visual of the tree.

        Parameters
        ----------
        depth : int, optional
            Maximum directory depth to display. ``None`` indicates no limit.
        hidden : bool
            If True, show hidden files and directories.

        """
        if not self.exists:
            raise OSError("Tree doesn't exist in the filesystem")

        tree = OrderedDict()
        rootdir = self.abspath.rstrip(os.sep)
        start = rootdir.rfind(os.sep) + 1
        for path, dirs, files in scandir.walk(rootdir):

            # sort files and directories so they show up in output sorted
            dirs.sort()
            files.sort()

            folders = ["{}/".format(x) for x in path[start:].split(os.sep)]
            parent = reduce(dict.get, folders[:-1], tree)

            # depth handling
            if depth and len(folders) == depth+1:
                parent[folders[-1]] = {}
                continue
            elif depth and len(folders) > depth+1:
                continue

            # filter out hidden files and directories, if desired
            if not hidden:
                outfiles = [file for file in files if file[0] != os.extsep]
                hidden_dirs = [d for d in dirs if d[0] == os.extsep]
                for d in hidden_dirs:
                    dirs.remove(d)
            else:
                outfiles = files

            subdir = OrderedDict.fromkeys(outfiles, {})
            parent[folders[-1]] = subdir

        tr = LeftAligned()
        print(tr(tree))

    def makedirs(self):
        """Make all directories along path that do not currently exist.

        Returns
        -------
        Tree
            This Tree.

        """
        makedirs(str(self.path), exist_ok=True)

        return self

    def make(self):
        """Make the directory if it doesn't exist. Equivalent to :meth:`makedirs`.

        Returns
        -------
        Tree
            This Tree.

        """
        self.makedirs()

        return self

    def sync(self, other, mode='upload', compress=True, checksum=True,
             backup=False, dry=False, include=None, exclude=None,
             overwrite=False, rsync_path='/usr/bin/rsync'):
        """Synchronize directories using rsync.

        Parameters
        ----------
        other: str or Tree
            Other end of the sync, can be either a path or another Tree.
        mode: str
            Either ``"upload"`` if uploading to  `other`, or ``"download"`` if
            downloading from `other`


        The other options are described in the
        :py:func:`datreant.rsync.rsync` documentation.

        """
        if isinstance(other, Tree):
            other = other.abspath

        if mode == 'download':
            source = other
            dest = self.abspath
        elif mode == 'upload':
            source = self.abspath
            dest = other
        else:
            raise ValueError("Sync mode can be only 'upload' or 'download'.")
        # Here we do some massaging before passing to the rsync function
        return rsync(source, dest, compress=compress, backup=backup,
                     dry=dry, include=include, checksum=checksum,
                     overwrite=overwrite, exclude=exclude,
                     rsync_path=rsync_path)


class _Loc(object):
    """Path accessor for Trees."""

    def __init__(self, tree):
        self._tree = tree

    def __getitem__(self, path):
        """Get Tree/Leaf at `path` relative to attached Tree.

        """
        return self._tree[path]


class _TreeLoc(_Loc):
    """Tree accessor for Trees."""

    def __getitem__(self, path):
        """Get Tree at `path` relative to attached Tree.

        """
        return Tree(self._tree[path].abspath)


class _LeafLoc(_Loc):
    """Leaf accessor for Trees."""

    def __getitem__(self, path):
        """Get Leaf at `path` relative to attached Tree.

        """
        return Leaf(self._tree[path].abspath)



# -*-encoding: utf-8 -*-
import sys
from PyQt5.QtWidgets import QApplication
from moldyn.ui.mainwindow import MoldynMainWindow


def main():
    app = QApplication(sys.argv)

    window = MoldynMainWindow()
    window.show()

    app.exec_()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'edit_model.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_EditModel(object):
    def setupUi(self, EditModel):
        EditModel.setObjectName("EditModel")
        EditModel.resize(493, 457)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(EditModel)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBox = QtWidgets.QGroupBox(EditModel)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.enableXPeriodicBoudariesLabel = QtWidgets.QLabel(self.groupBox)
        self.enableXPeriodicBoudariesLabel.setObjectName("enableXPeriodicBoudariesLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.enableXPeriodicBoudariesLabel)
        self.enableXPeriodicBoudariesCheckBox = QtWidgets.QCheckBox(self.groupBox)
        self.enableXPeriodicBoudariesCheckBox.setObjectName("enableXPeriodicBoudariesCheckBox")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.enableXPeriodicBoudariesCheckBox)
        self.enableYPeriodicBoudariesLabel = QtWidgets.QLabel(self.groupBox)
        self.enableYPeriodicBoudariesLabel.setObjectName("enableYPeriodicBoudariesLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.enableYPeriodicBoudariesLabel)
        self.enableYPeriodicBoudariesCheckBox = QtWidgets.QCheckBox(self.groupBox)
        self.enableYPeriodicBoudariesCheckBox.setObjectName("enableYPeriodicBoudariesCheckBox")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.enableYPeriodicBoudariesCheckBox)
        self.gridLayout_3.addLayout(self.formLayout, 0, 0, 1, 1)
        self.verticalLayout.addWidget(self.groupBox)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(EditModel)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_2.addWidget(self.buttonBox)
        self.enableXPeriodicBoudariesLabel.setBuddy(self.enableXPeriodicBoudariesCheckBox)
        self.enableYPeriodicBoudariesLabel.setBuddy(self.enableYPeriodicBoudariesCheckBox)

        self.retranslateUi(EditModel)
        QtCore.QMetaObject.connectSlotsByName(EditModel)

    def retranslateUi(self, EditModel):
        _translate = QtCore.QCoreApplication.translate
        EditModel.setWindowTitle(_translate("EditModel", "Edit Model"))
        self.groupBox.setTitle(_translate("EditModel", "Spatial configuration"))
        self.enableXPeriodicBoudariesLabel.setText(_translate("EditModel", "Enable X periodic boudaries"))
        self.enableYPeriodicBoudariesLabel.setText(_translate("EditModel", "Enable Y periodic boudaries"))




from pyparsing import (CaselessLiteral, Word, quotedString,
                       removeQuotes, operatorPrecedence, opAssoc, stringEnd,
                       ParseException)

__all__ = ['parse_selection']


class UnaryOperation(object):
    def __init__(self, t):
        self.op, self.a = t[0]


class BinaryOperation(object):
    def __init__(self, t):
        self.op = t[0][1]
        self.operands = t[0][0::2]


class SearchAnd(BinaryOperation):
    def generate_tag_expr(self):
        return list(oper.generate_tag_expr() for oper in self.operands)

    def __repr__(self):
        return "AND:(%s)" % (",".join(str(oper) for oper in self.operands))


class SearchOr(BinaryOperation):
    def generate_tag_expr(self):
        return tuple(oper.generate_tag_expr() for oper in self.operands)

    def __repr__(self):
        return "OR:(%s)" % (",".join(str(oper) for oper in self.operands))


class SearchNot(UnaryOperation):
    def generate_tag_expr(self):
        exps = self.a.generate_tag_expr()
        if isinstance(exps, list):
            return [{e} for e in exps]
        elif isinstance(exps, tuple):
            return tuple({e} for e in exps)
        else:
            return {exps}

    def __repr__(self):
        return "NOT:(%s)" % str(self.a)


class SearchTerm(object):
    def __init__(self, tokens):
        self.term = tokens[0]

    def generate_tag_expr(self):
        return "{}".format(self.term)

    def __repr__(self):
        return self.term


# define the grammar
and_ = CaselessLiteral("and")
or_ = CaselessLiteral("or")
not_ = CaselessLiteral("not")
allowed_chars = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'*+,-./:;<=>?@[]^_`{|}~'
# first remove matching strings and then parse for printable characters
searchTerm = quotedString.setParseAction(removeQuotes) | Word(allowed_chars)
searchTerm.setParseAction(SearchTerm)
searchExpr = operatorPrecedence(searchTerm, [
    (not_, 1, opAssoc.RIGHT, SearchNot),
    (and_, 2, opAssoc.LEFT, SearchAnd),
    (or_, 2, opAssoc.LEFT, SearchOr),
])
Parser = searchExpr + stringEnd


def parse_selection(sel):
    """
    Parse a tag selection string and convert it to the default list/tuple/set
    tag selections. If the selection can't be parsed the original selection
    is returned.

    Parameters
    ----------
    sel : str
        selection string

    Returns
    -------
    list/tuple/set representation used to filter tags

    Example
    -------
    >>> parse_selection('food and drinks')
    ['food', 'drinks']
    >>> parse_selection('free beer')
    'free beer'
    """
    try:
        return Parser.parseString(sel)[0].generate_tag_expr()
    except ParseException:
        return sel



def _float(w, default=None):
    if default==None:
        return float(w.text().replace(",", "."))
    else:
        try:
            return _float(w)
        except:
            w.setText(str(default))
            return default



"""Stress tests for file locks.

"""

import multiprocessing as mp
import pytest

from datreant import Treant


@pytest.fixture
def treant(tmpdir):
    with tmpdir.as_cwd():
        t = Treant('sprout')
    return t


def pokefile(treantpath, string):
    """Add a number of tags to a Treant."""
    treant = Treant(treantpath)
    treant.tags.add(* ["{}_{}".format(string, i) for i in range(100)])


def test_death_by_1000_pokes(treant):
    pool = mp.Pool(processes=4)
    for i in range(10):
        pool.apply_async(pokefile, args=(treant.abspath, "run_{}".format(i)))
    pool.close()
    pool.join()

    assert len(treant.tags) == 1000


def init_treant(tmpdir, tags):
    with tmpdir.as_cwd():
        tf = Treant('sprout', tags=tags)
    return tf


def test_init_treant(tmpdir):
    pool = mp.Pool(processes=4)
    num = 73

    # TODO: eventually want this to work without initing the treant
    # here
    init_treant(tmpdir, ['bark'])
    for i in range(num):
        pool.apply_async(init_treant, args=(tmpdir, ['run_{}'.format(i)]))
    pool.close()
    pool.join()

    with tmpdir.as_cwd():
        tf = Treant('sprout')

    assert len(tf.tags) == num + 1



import os


def makedirs(path, mode=0o777, exist_ok=False):
    """Make directories and all parents necessary.

    :Arguments:
        *path*
            directory path to make
    """
    try:
        os.makedirs(str(path), mode=mode)
    except OSError as e:
        # let slide errors that include directories already existing, but
        # catch others
        if e.errno == 17 and exist_ok:
            pass
        else:
            raise e



"""
Metadata components (tags, categories) give texture to directories, allowing
for filtering and selection.
"""
import os
import itertools
import functools
from six import string_types
from six.moves import range
from collections import defaultdict

from fuzzywuzzy import process

from .state import JSONFile
from .selectionparser import parse_selection


class Metadata(object):

    # TODO; raise meaningful exceptions when file can't be read, written, read
    # and nonexistent, missing, or treantdir missing

    def __init__(self, tree):
        self._tree = tree

    @property
    def _write(self):
        self._statefile = JSONFile(os.path.join(self._tree._treantdir,
                                                self._statefilename),
                                   init_state=self._init_state)

        return self._statefile.write()

    @property
    def _read(self):
        self._statefile = JSONFile(os.path.join(self._tree._treantdir,
                                                self._statefilename),
                                   init_state=self._init_state)

        return self._statefile.read()


@functools.total_ordering
class Tags(Metadata):
    """Interface to tags.

    """
    _statefilename = 'tags.json'

    @staticmethod
    def _init_state(jsonfile):
        """Used solely for initializing JSONFile state for storing tag
        information.

        """
        jsonfile._state = []

    def __repr__(self):
        return "<Tags({})>".format(self._list())

    def __str__(self):
        tags = self._list()
        agg = "Tags"
        majsep = "="
        seplength = len(agg)

        if not tags:
            out = "No Tags"
        else:
            out = agg + '\n'
            out = out + majsep * seplength + '\n'
            for i in range(len(tags)):
                out = out + "'{}'\n".format(tags[i])
        return out

    def __getitem__(self, value):
        # check if we might have a string to parse into a selection object
        if isinstance(value, string_types):
            value = parse_selection(value)
        return self._getselection(value)

    def _getselection(self, value):
        """get tags according to a selection
        """
        with self._read:
            if isinstance(value, list):
                # a list of tags gives only members with ALL the tags
                fits = all(self._getselection(item) for item in value)
            elif isinstance(value, tuple):
                # a tuple of tags gives members with ANY of the tags
                fits = any(self._getselection(item) for item in value)
            if isinstance(value, set):
                # a set of tags gives only members WITHOUT ALL the tags
                # can be used for `not`, basically
                fits = not all(self._getselection(item) for item in value)
            elif isinstance(value, string_types):
                fits = value in self

            return fits

    def __iter__(self):
        return self._list().__iter__()

    def __len__(self):
        return len(self._list())

    def __eq__(self, other):
        if isinstance(other, (Tags, set, list)):
            return set(self) == set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def __lt__(self, other):
        if isinstance(other, (Tags, set, list)):
            return set(self) < set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def __sub__(self, other):
        """Return a set giving the Tags in `a` that are not in `b`.

        """
        if isinstance(other, (AggTags, Tags, set, list)):
            return set(self) - set(other)
        else:
            raise TypeError("Operands must be AggTags, tags, a set, or list.")

    def __rsub__(self, other):
        """Return a set giving the Tags in `a` that are not in `b`.

        """
        if isinstance(other, (AggTags, Tags, set, list)):
            return set(other) - set(self)
        else:
            raise TypeError("Operands must be AggTags, tags, a set, or list.")

    def __or__(self, other):
        """Return a set giving the union of Tags `a` and `b`.

        """
        if isinstance(other, (Tags, set, list)):
            return set(self) | set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def __ror__(self, other):
        """Return a set giving the union of Tags `a` and `b`.

        """
        if isinstance(other, (Tags, set, list)):
            return set(self) | set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def __and__(self, other):
        """Return a set giving the intersection of Tags `a` and `b`.

        """
        if isinstance(other, (Tags, set, list)):
            return set(self) & set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def __rand__(self, other):
        """Return a set giving the intersection of Tags `a` and `b`.

        """
        if isinstance(other, (Tags, set, list)):
            return set(self) & set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def __xor__(self, other):
        """Return a set giving the symmetric difference of Tags
        `a` and `b`.

        """
        if isinstance(other, (Tags, set, list)):
            return set(self) ^ set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def __rxor__(self, other):
        """Return a set giving the symmetric difference of Tags
        `a` and `b`.

        """
        if isinstance(other, (Tags, set, list)):
            return set(self) ^ set(other)
        else:
            raise TypeError("Operands must be tags, a set, or list.")

    def _list(self):
        """Get all tags for the Treant as a list.

        :Returns:
            *tags*
                list of all tags
        """
        with self._read:
            tags = self._statefile._state

        tags.sort()
        return tags

    def add(self, *tags):
        """Add any number of tags to the Treant.

        Tags are individual strings that serve to differentiate Treants from
        one another. Sometimes preferable to categories.

        Parameters
        ----------
        tags : str or list
            Tags to add. Must be strings or lists of strings.

        """
        outtags = list()
        for tag in tags:
            if isinstance(tag, (list, set, tuple)):
                outtags.extend(tag)
            else:
                outtags.append(tag)

        with self._write:
            # ensure tags are unique (we don't care about order)
            # also they must be strings
            _outtags = []
            for tag in outtags:
                if not isinstance(tag, string_types):
                    raise ValueError("Only string can be added as tags. Tried "
                                     "to add '{}' which is '{}'".format(
                                         tag, type(tag)))
                _outtags.append(tag)
            outtags = set(_outtags)

            # remove tags already present in metadata from list
            outtags = outtags.difference(set(self._statefile._state))

            # add new tags
            self._statefile._state.extend(outtags)

    def remove(self, *tags):
        """Remove tags from Treant.

        Any number of tags can be given as arguments, and these will be
        deleted.

        :Arguments:
            *tags*
                Tags to delete.
        """
        with self._write:
            # remove redundant tags from given list if present
            tags = set([str(tag) for tag in tags])
            for tag in tags:
                # remove tag; if not present, continue anyway
                try:
                    self._statefile._state.remove(tag)
                except ValueError:
                    pass

    def clear(self):
        """Remove all tags from Treant.

        """
        with self._write:
            self._statefile._state = list()

    def fuzzy(self, tag, threshold=80):
        """Get a tuple of existing tags that fuzzily match a given one.

        Parameters
        ----------
        tags : str or list
            Tag or tags to get fuzzy matches for.
        threshold : int
            Lowest match score to return. Setting to 0 will return every tag,
            while setting to 100 will return only exact matches.

        Returns
        -------
        matches : tuple
            Tuple of tags that match.
        """
        if isinstance(tag, string_types):
            tags = [tag]
        else:
            tags = tag

        matches = []

        for tag in tags:
            matches += [i[0] for i in process.extract(tag, self, limit=None)
                        if i[1] > threshold]

        return tuple(matches)


class Categories(Metadata):
    """Interface to categories.

    """
    _statefilename = 'categories.json'

    @staticmethod
    def _init_state(jsonfile):
        """Used solely for initializing JSONFile state for storing category
        information.

        """
        jsonfile._state = {}

    def __repr__(self):
        return "<Categories({})>".format(self._dict())

    def __str__(self):
        categories = self._dict()
        agg = "Categories"
        majsep = "="
        seplength = len(agg)

        if not categories:
            out = "No Categories"
        else:
            out = agg + '\n'
            out = out + majsep * seplength + '\n'
            for key in categories.keys():
                out = out + "'{}': '{}'\n".format(key, categories[key])
        return out

    def __getitem__(self, keys):
        """Get values for given `keys`.

        If `keys` is a string, the single value for that string is returned.

        If `keys` is a list of keys, the values for each key are returned in a
        list, in order by the given keys.

        if `keys` is a set of keys, a dict with the keys as keys and values as
        values is returned.

        Parameters
        ----------
        keys : str, list, set
            Key(s) of value to return.

        Returns
        -------
        values : str, int, float, bool, list, or dict
            Value(s) corresponding to given key(s).

        """
        categories = self._dict()

        if isinstance(keys, (int, float, string_types, bool)):
            return categories[keys]
        elif isinstance(keys, list):
            return [categories[key] for key in keys]
        elif isinstance(keys, set):
            return {key: categories[key] for key in keys}
        else:
            raise TypeError("Key must be a string, list of strings, or set"
                            " of strings.")

    def __setitem__(self, key, value):
        """Set value at given key.

        Parameters
        ----------
        key : str
            Key of value to set.
        value : str, int, float, bool
            Value to set for given key.

        """
        outdict = {key: value}
        self.add(outdict)

    def __delitem__(self, category):
        """Remove category from Treant.

        """
        self.remove(category)

    def __eq__(self, other):
        if isinstance(other, (Categories, dict)):
            return dict(self) == dict(other)
        else:
            raise TypeError("Operands must be categories or dicts.")

    def __req__(self, other):
        if isinstance(other, (Categories, dict)):
            return dict(self) == dict(other)
        else:
            raise TypeError("Operands must be categories or dicts.")

    def __iter__(self):
        return self._dict().__iter__()

    def __len__(self):
        return len(self._dict())

    def _dict(self):
        """Get all categories for the Treant as a dictionary.

        :Returns:
            *categories*
                dictionary of all categories

        """
        with self._read:
            return self._statefile._state

    def add(self, categorydict=None, **categories):
        """Add any number of categories to the Treant.

        Categories are key-value pairs that serve to differentiate Treants from
        one another. Sometimes preferable to tags.

        If a given category already exists (same key), the value given will
        replace the value for that category.

        Keys must be strings.

        Values may be ints, floats, strings, or bools. ``None`` as a value
        will delete the existing value for the key, if present, and is
        otherwise not allowed.

        Parameters
        ----------
        categorydict : dict
            Dict of categories to add; keys used as keys, values used as
            values.
        categories : dict
            Categories to add. Keyword used as key, value used as value.

        """
        outcats = dict()
        if isinstance(categorydict, (dict, Categories)):
            outcats.update(categorydict)
        elif categorydict is None:
            pass
        else:
            raise TypeError("Invalid arguments; non-keyword"
                            " argument must be dict")

        outcats.update(categories)

        with self._write:
            for key, value in outcats.items():
                if not isinstance(key, string_types):
                    raise TypeError("Keys must be strings.")

                if (isinstance(value, (int, float, string_types, bool))):
                    self._statefile._state[key] = value
                elif value is None:
                    # delete if present
                    if key in self._statefile._state:
                        del self._statefile._state[key]
                    else:
                        raise ValueError("Cannot set to 'None'")
                else:
                    raise TypeError("Values must be ints, floats,"
                                    " strings, or bools.")

    def remove(self, *categories):
        """Remove categories from Treant.

        Any number of categories (keys) can be given as arguments, and these
        keys (with their values) will be deleted.

        Parameters
        ----------
        categories : str
                Categories to delete.

        """
        with self._write:
            for key in categories:
                # continue even if key not already present
                self._statefile._state.pop(key, None)

    def clear(self):
        """Remove all categories from Treant.

        """
        with self._write:
            self._statefile._state = dict()

    def keys(self):
        """Get category keys.

        :Returns:
            *keys*
                keys present among categories
        """
        with self._read:
            return self._statefile._state.keys()

    def values(self):
        """Get category values.

        :Returns:
            *values*
                values present among categories
        """
        with self._read:
            return self._statefile._state.values()


class AggMetadata(object):

    def __init__(self, collection):
        self._collection = collection


@functools.total_ordering
class AggTags(AggMetadata):
    """Interface to aggregated tags.

    """

    def __init__(self, collection):
        super(AggTags, self).__init__(collection)

    def __repr__(self):
        return "<AggTags({})>".format(list(self.all))

    def __str__(self):
        tags = list(self.all)
        agg = "Tags"
        majsep = "="
        seplength = len(agg)

        if not tags:
            out = "No Tags"
        else:
            out = agg + '\n'
            out = out + majsep * seplength + '\n'
            for i in xrange(len(tags)):
                out = out + "'{}'\n".format(tags[i])
        return out

    def __iter__(self):
        return self.all.__iter__()

    def __len__(self):
        return len(self.all)

    def __getitem__(self, value):
        return [member.tags[value] for member in self._collection]

    def __eq__(self, other):
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) == set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __lt__(self, other):
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) < set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __sub__(self, other):
        """Return a set giving the Tags in `a` that are not in `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) - set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __rsub__(self, other):
        """Return a set giving the Tags in `a` that are not in `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(other) - set(self)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __or__(self, other):
        """Return a set giving the union of Tags `a` and `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) | set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __ror__(self, other):
        """Return a set giving the union of Tags `a` and `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) | set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __and__(self, other):
        """Return a set giving the intersection of Tags `a` and `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) & set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __rand__(self, other):
        """Return a set giving the intersection of Tags `a` and `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) & set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __xor__(self, other):
        """Return a set giving the symmetric difference of Tags
        `a` and `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) ^ set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    def __rxor__(self, other):
        """Return a set giving the symmetric difference of Tags
        `a` and `b`.

        """
        if isinstance(other, (AggTags, Tags, set)):
            return set(self) ^ set(other)
        else:
            raise TypeError("Operands must be AggTags, Tags, or a set.")

    @property
    def any(self):
        """Set of tags present among at least one Treant in collection.

        """
        tags = [set(member.tags) for member in self._collection]
        out = set.union(*tags)

        return out

    @property
    def all(self):
        """Set of tags present among all Treants in collection.

        """
        tags = [set(member.tags) for member in self._collection]
        out = set.intersection(*tags)

        return out

    def add(self, *tags):
        """Add any number of tags to each Treant in collection.

        :Arguments:
           *tags*
              Tags to add. Must be strings or lists of strings.

        """
        for member in self._collection:
            member.tags.add(*tags)

    def remove(self, *tags):
        """Remove tags from each Treant in collection.

        Any number of tags can be given as arguments, and these will be
        deleted.

        :Arguments:
            *tags*
                Tags to delete.
        """
        for member in self._collection:
            member.tags.remove(*tags)

    def clear(self):
        """Remove all tags from each Treant in collection.

        """
        for member in self._collection:
            member.tags.clear()

    def fuzzy(self, tag, threshold=80, scope='all'):
        """Get a tuple of existing tags that fuzzily match a given one.

        Parameters
        ----------
        tag : str or list
            Tag or tags to get fuzzy matches for.
        threshold : int
            Lowest match score to return. Setting to 0 will return every tag,
            while setting to 100 will return only exact matches.
        scope : {'all', 'any'}
            Tags to use. 'all' will use only tags found within all Treants in
            collection, while 'any' will use tags found within at least one
            Treant in collection.

        Returns
        -------
        matches : tuple
            Tuple of tags that match.
        """
        if isinstance(tag, string_types):
            tags = [tag]
        else:
            tags = tag

        if scope == 'all':
            choices = self.all
        elif scope == 'any':
            choices = self.any
        else:
            raise ValueError("Scope can only be 'any' or 'all'")

        matches = []

        for tag in tags:
            matches += [i[0] for i in process.extract(tag, choices, limit=None)
                        if i[1] > threshold]

        return tuple(matches)

    def filter(self, tag):
        """Filter Treants matching the given tag expression from a Bundle.

        Parameters
        ----------
        tag : str or list
            Tag or tags to filter Treants.

        Returns
        -------
        Bundle
            Bundle of Treants matching the given tag expression.
        """
        matches = self._collection.tags[tag]
        return self._collection[matches]


class AggCategories(AggMetadata):
    """Interface to categories.

    """

    def __init__(self, collection):
        super(AggCategories, self).__init__(collection)

    def __repr__(self):
        return "<AggCategories({})>".format(self.all)

    def __str__(self):
        categories = self.all
        agg = "Categories"
        majsep = "="
        seplength = len(agg)

        if not categories:
            out = "No Categories"
        else:
            out = agg + '\n'
            out = out + majsep * seplength + '\n'
            for key, value in categories.items():
                out = out + "'{}': '{}'\n".format(key, value)
        return out

    def __getitem__(self, keys):
        """Get values for a given key, list of keys, or set of keys.

        If `keys` is a string specifying one key (for a single category),
        return a list of the values among all Treants (in this collection) for
        that category.

        If `keys` is a list of keys, return a list of lists whose order
        corresponds to the order of the elements in `keys`. Each element in
        `keys` is a key specifying a category; each element in the output is
        a list of the values among all Treants (in this collection) for  the
        category specified by the respective key in `keys`.

        If `keys` is a set of keys, return a dict of lists whose keys are the
        same as those provided in `keys`; the value corresponding to each key
        in the output is a list of values among all Treants (in this
        collection) for the category corresponding to that key.

        Parameters
        ----------
        keys : str, list, set
            Valid key(s) of Categories in this collection.

        Returns
        -------
        list, list of lists, dict of lists
            Values for the (single) specified category when `keys` is str.

        """
        if keys is None:
            return None

        members = self._collection
        if isinstance(keys, (int, float, string_types, bool)):
            k = keys
            return [m.categories[k] if k in m.categories else None
                    for m in members]
        elif isinstance(keys, list):
            return [[m.categories[k] if k in m.categories else None
                    for m in members]
                    for k in keys]
        elif isinstance(keys, set):
            return {k: [m.categories[k] if k in m.categories else None
                    for m in members]
                    for k in keys}
        else:
            raise TypeError("Key must be a string, list of strings, or set"
                            " of strings.")

    def __setitem__(self, key, values):
        """Set the value of categories for each Treant in the collection.

        If `values` is not a sequence and is a valid category type (int,
        string_types, bool, float), then it is broadcasted over all members of
        the collection for the category specified by `key`.

        If `values` is a sequence, it must have the same length as the number
        of members in the collection so that, for each member, the value
        assigned to its category (specified by `key`) is the element in
        `values` whose index matches the index of that member in the
        collection.

        Parameters
        ----------
        key : str
            Valid key for the category whose value should be set to *values*.
        values : str, int, float, bool, list, tuple
            Value(s) for the category specified by `key`.
        """
        if values is None:
            return

        members = self._collection
        if isinstance(values, (int, float, string_types, bool)):
            for m in members:
                m.categories.add({key: values})
        elif isinstance(values, (list, tuple)):
            if len(values) != len(members):
                raise ValueError("Values must be a list of the same length as"
                                 " the number of members in the collection.")
            for m, v in zip(members, values):
                m.categories[key] = v

    def __delitem__(self, category):
        """Remove *category* from each Treant in collection.

        """
        for member in self._collection:
            member.categories.remove(category)

    def __iter__(self):
        """Iterator over Categories common to all Treants in collection.

        """
        return self.all.__iter__()

    def __len__(self):
        """The number of Categories common to all Treants in collection.

        """
        return len(self.all)

    @property
    def any(self):
        """Get categories present among at least one Treant in collection.

        Returns
        -------
        dict
            All unique Categories among members.
        """
        keys = [set(member.categories.keys()) for member in self._collection]
        keys = set.union(*keys)

        return {k: [m.categories[k] if k in m.categories else None
                for m in self._collection]
                for k in keys}

    @property
    def all(self):
        """Get categories common to all Treants in collection.

        Returns
        -------
        dict
            Categories common to all members.
        """
        keys = [set(member.categories.keys()) for member in self._collection]
        keys = set.intersection(*keys)

        return {k: [m.categories[k] if k in m.categories else None
                for m in self._collection]
                for k in keys}

    def add(self, categorydict=None, **categories):
        """Add any number of categories to each Treant in collection.

        Categories are key-value pairs that serve to differentiate Treants from
        one another. Sometimes preferable to tags.

        If a given category already exists (same key), the value given will
        replace the value for that category.

        Keys must be strings.

        Values may be ints, floats, strings, or bools. ``None`` as a value
        will not the existing value for the key, if present.

        Parameters
        ----------
        categorydict : dict
            Dict of categories to add; keys used as keys, values used as
            values.
        categories
            Categories to add. Keyword used as key, value used as value.

        """
        for member in self._collection:
            member.categories.add(categorydict, **categories)

    def remove(self, *categories):
        """Remove categories from Treant.

        Any number of categories (keys) can be given as arguments, and these
        keys (with their values) will be deleted.

        Parameters
        ----------
        categories : str
            Categories to delete.
        """
        for member in self._collection:
            member.categories.remove(*categories)

    def clear(self):
        """Remove all categories from all Treants in collection.

        """
        for member in self._collection:
            member.categories.clear()

    def keys(self, scope='all'):
        """Get the keys present among Treants in collection.

        Parameters
        ----------
        scope : {'all', 'any'}
            Keys to return. 'all' will return only keys found within all
            Treants in the collection, while 'any' will return keys found
            within at least one Treant in the collection.

        Returns
        -------
        keys : list
            Present keys.

        """
        keys = [set(member.categories.keys()) for member in self._collection]

        if scope == 'all':
            out = set.intersection(*keys)
        elif scope == 'any':
            out = set.union(*keys)
        else:
            raise ValueError("Scope must be either 'all' or 'any'")

        return list(out)

    def values(self, scope='all'):
        """Get the category values for all Treants in collection.

        Parameters
        ----------
        scope : {'all', 'any'}
            Keys to return. 'all' will return only keys found within all
            Treants in the collection, while 'any' will return keys found
            within at least one Treant in the collection.

        Returns
        -------
        values : list
            A list of values for each Treant in the collection is returned for
            each key within the given `scope`. The value lists are given in
            the same order as the keys from ``AggCategories.keys``.

        """
        keys = self.keys(scope=scope)
        return self[keys]

    def groupby(self, keys):
        """Return groupings of Treants based on values of Categories.

        If a single category is specified by `keys` (`keys` is neither a list
        nor a set of category names), returns a dict of Bundles whose (new)
        keys are the values of the category specified by `keys`; the
        corresponding Bundles are groupings of members in the collection having
        the same category values (for the category specied by `keys`).

        If `keys` is a list of keys, returns a dict of Bundles whose
        (new) keys are tuples of category values. The corresponding Bundles
        contain the members in the collection that have the same set of
        category values (for the categories specified by `keys`); members in
        each Bundle will have all of the category values specified by the tuple
        for that Bundle's key.

        Parameters
        ----------
        keys : str, list
            Valid key(s) of categories in this collection.

        Returns
        -------
        dict
            Bundles of members by category values.
        """
        from .collections import Bundle

        if keys is None:
            return None

        members = self._collection
        if isinstance(keys, (string_types)):
            catvals = members.categories[keys]
            groupkeys = [v for v in catvals if v is not None]
            groups = {k: Bundle() for k in groupkeys}

            k = keys
            gen = ((m, m.categories[k]) for m in members if
                   k in m.categories and m.categories[k] in groupkeys)
            for m, catval in gen:
                groups[catval] += m
        # Note: redundant code in if/elif block can be consolidated in future
        elif isinstance(keys, list):
            catvals = list(zip(*members.categories[keys]))
            groupkeys = [v for v in catvals if None not in v]
            groups = {k: Bundle() for k in groupkeys}

            gen = ((i, m) for i, m in enumerate(members) if
                   set(keys) <= set(m.categories) and
                   tuple([m.categories[k] for k in keys]) in groupkeys)
            for i, m in gen:
                groups[catvals[i]] += m

        else:
            raise TypeError("Keys must be a string or a list of"
                            " strings")
        return groups



import matplotlib
import matplotlib.pyplot as plt
from PyQt5.QtWidgets import (QDialog, QLineEdit, QApplication, QGridLayout,
                             QPushButton, QLabel)
from matplotlib.widgets import Button
import sys


class MultiInputDialog(QDialog):
    def __init__(self, labels, vals, format_dict, parent=None):
        super().__init__(parent)
        self.vals = list(vals)
        self.labels = []
        self.edits = []
        for key, value in zip(labels, vals):
            if key in format_dict:
                value = format_dict[key].format(value)
            else:
                value = str(value)
            self.labels.append(QLabel(key))
            edit = QLineEdit()
            edit.setText(value)
            self.edits.append(edit)
        
        self.accept_button = QPushButton("Accept")
        self.accept_button.clicked.connect(self.accept)
        # Create layout and add widgets

        layout = QGridLayout()
        for i, label, edit in zip(range(len(self.labels)), self.labels, self.edits):
            layout.addWidget(label, i, 0)
            layout.addWidget(edit, i, 1)
        layout.addWidget(self.accept_button, i+1, 1)

        # Set dialog layout

        self.setLayout(layout)

    def exec(self) -> int:
        self.show()
        self.edits[0].setFocus()
        return super().exec()

    def result(self):
        print("result", super().result())
        if super().result():
            for i, edit in enumerate(self.edits):
                self.vals[i] = edit.text()
        return self.vals


class DraggableLine:
    def __init__(self, x_data, y_data, picker, x_label, y_label, axis, y_min=None, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.y_min = y_min
        self.picker = picker
        self.fig = plt.figure()
        ax = self.fig.add_subplot(111)
        plt.subplots_adjust(bottom=0.2)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.grid()
        self.x_label = x_label
        self.y_label = y_label
        ax.axis(axis)
        self.line, = ax.plot(x_data, y_data, *args, picker=picker, **kwargs)
        self.axnew = plt.axes([0.65, 0.02, 0.25, 0.05])
        self.bnew = Button(self.axnew, 'New point (ctrl)')
        self.axdel = plt.axes([0.39, 0.02, 0.25, 0.05])
        self.bdel = Button(self.axdel, 'Delete point (alt)')
        self.axsetaxis = plt.axes([0.13, 0.02, 0.25, 0.05])
        self.bsetaxis = Button(self.axsetaxis, 'Set axis')
        self.bnew.on_clicked(self.new)
        self.bdel.on_clicked(self.del_point)
        self.bsetaxis.on_clicked(self.set_axis)
        self.press = None
        self.obj = None
        self.init_connect()

    def _redraw(self, x_data, y_data):
        ax = self.line.axes
        axis = ax.axis()
        ax.cla()
        self.line, = ax.plot(x_data, y_data, *self.args, picker=self.picker, **self.kwargs)
        ax.axis(axis)
        ax.grid()
        ax.set_xlabel(self.x_label)
        ax.set_ylabel(self.y_label)


    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.line.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidpick = self.line.figure.canvas.mpl_connect(
            'pick_event', self.on_pick)
        self.cidrelease = self.line.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.line.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def init_connect(self):
        self.new_pending = False
        self.new_key_press = False
        self.del_key_press = False
        self.del_pending = False
        self.cidaxes_enter = self.line.figure.canvas.mpl_connect(
            'axes_enter_event', self.on_enter)
        self.cidaxes_leave = self.line.figure.canvas.mpl_connect(
            'axes_leave_event', self.on_leave)
        self.cidkpress = self.line.figure.canvas.mpl_connect(
            'key_press_event', self.key_press)
        self.cidkrelease = self.line.figure.canvas.mpl_connect(
            'key_release_event', self.key_press)

    def key_press(self, event):
        if event.key == 'control':
            self.new_key_press = not self.new_key_press
            print("New point :", self.new_key_press)
        elif event.key ==  "alt":
            self.del_key_press = not self.del_key_press
            print("Del point :", self.del_key_press)

    def key_release(self, event):
        print('release', event.key)
        if event.key == 'control':
            self.new_key_press = False
        elif event.key ==  "alt":
            self.del_key_press = False

    def new(self, event):
        self.new_pending = not self.new_pending

    def del_point(self, event):
        self.del_pending = not self.del_pending
        
    def on_enter(self, event):
        self.connect()

    def on_leave(self, event):
        self.disconnect()
    
    def on_press(self, event):
        '''on button press we will see if the mouse is over us
            and store some data'''
        if event.inaxes != self.line.axes:
            print('!=axes')
            return

        # create new point
        if self.new_pending or self.new_key_press:
            #print("new point")
            x, y = event.xdata, event.ydata
            x_data = list(self.line.get_xdata())
            y_data = list(self.line.get_ydata())
            i = 0
            while i<len(x_data) and x > x_data[i]:
                i += 1
            x_data = x_data[:i] + [x] + x_data[i:]
            if self.y_min is None:
                y_data = y_data[:i] + [y] + y_data[i:]
            else:
                y_data = y_data[:i] + [max(y, self.y_min)] + y_data[i:]
            self._redraw(x_data, y_data)
            self.on_release(event)
            self.new_pending = False
            return            

        # double-click edit
        if event.dblclick and self.obj is not None:
            #print("double clique")
            x0, y0, ind = self.obj
            d = MultiInputDialog(('x', 'y'), (x0, y0), {'x': '{:.3e}', 'y': '{:.3e}'})
            d.exec()
            x, y = map(float, d.result())
            x_data = self.line.get_xdata()
            y_data = self.line.get_ydata()
            if self.y_min is None:
                x_data[ind], y_data[ind] = x, y
            else:
                x_data[ind], y_data[ind] = x, max(y, self.y_min)
            self.line.set_xdata(x_data)
            self.line.set_ydata(y_data)
            self.on_release(event)
            return
        self.press = event.xdata, event.ydata

    def on_pick(self, event):
        '''on button press we will see if the mouse is over us
            and store some data'''
        #if event.inaxes != self.rect.axes: return
        thisline = event.artist
        xdata = thisline.get_xdata()
        ydata = thisline.get_ydata()
        ind = event.ind
        points = tuple(zip(xdata[ind], ydata[ind]))
        #print('onpick points:', points)
        #print('event contains', ind)

        # delete points
        if self.del_key_press or self.del_pending:
            xdata = list(xdata)
            ydata = list(ydata)
            del xdata[ind[0]], ydata[ind[0]]
            self._redraw(xdata, ydata)
            self.on_release(event)
            self.del_pending = False
            return
        
        self.obj = *points[-1], ind

    def on_motion(self, event):
        '''on motion we will move the rect
            if the mouse is over us'''
        if self.press is None or self.obj is None: return
        if event.inaxes != self.line.axes:
            print('!=axes')
            return
        x0, y0, ind = self.obj
        xpress, ypress = self.press
        dx = event.xdata - xpress
        dy = event.ydata - ypress
        #print('x0=%f, xpress=%f, event.xdata=%f, dx=%f, x0+dx=%f' %
        #      (x0, xpress, event.xdata, dx, x0+dx))
        x_data = self.line.get_xdata()
        y_data = self.line.get_ydata()
        x_data[ind] = x0 + dx
        y_data[ind] = y0 + dy
        self.line.set_xdata(x_data)
        self.line.set_ydata(y_data)

        self.line.figure.canvas.draw()


    def on_release(self, event):
        '''on release we reset the press data'''
        self.press = None
        self.obj = None
        self.line.figure.canvas.draw()

    def set_axis(self, event):
        self.key_disconnect()
        ax = self.line.axes
        axis = ax.axis()
        d = MultiInputDialog(('x', "x'", 'y', "y'"), axis,
                             {'x':'{:.3e}', "x'":'{:.3e}', 'y':'{:.3e}', "y'":'{:.3e}'})
        d.exec()
        axis = list(map(float, d.result()))
        ax.axis(axis)
        self.init_connect()

    def key_disconnect(self):
        #print("disconnect")
        self.line.figure.canvas.mpl_disconnect(self.cidkpress)
        self.line.figure.canvas.mpl_disconnect(self.cidkrelease)
        self.line.figure.canvas.mpl_disconnect(self.cidaxes_enter)
        self.line.figure.canvas.mpl_disconnect(self.cidaxes_leave)
        self.disconnect()


    def disconnect(self):
        '''disconnect all the stored connection ids'''
        self.line.figure.canvas.mpl_disconnect(self.cidpress)
        self.line.figure.canvas.mpl_disconnect(self.cidpick)
        self.line.figure.canvas.mpl_disconnect(self.cidrelease)
        self.line.figure.canvas.mpl_disconnect(self.cidmotion)


def main(queue):
    """


    The queue must contain a list containing the following elements in order:
        - axis : (tuple) a four element tuple defining the axis of the figure
        - x_data : (1d-array) the data for the x axis
        - y_data : (1d-array) the data for the y axis
        - xlabel : (str) the label for the x axis
        - ylabel : (str) the label for the y axis
    Parameters
    ----------
    queue

    Returns
    -------

    """
    #app = QApplication(sys.argv)
    axis, x_data, y_data, xlabel, ylabel, y_min = queue.get()
    dr = DraggableLine(x_data, y_data, 5, xlabel, ylabel, axis, y_min, 'o--', color='r')
    plt.show()
    queue.put(dr.line.get_data())


if __name__ == "__main__":
    main()



"""Parser tests for Treants.

"""

from datreant.selectionparser import parse_selection

import pytest


@pytest.mark.parametrize('expr', ('foo bar', 'foo'))
def test_names(expr):
    assert parse_selection(expr) == expr


@pytest.mark.parametrize("expr, typename", (("foo and bar", list),
                                            ("foo or bar", tuple), ("not foo",
                                                                    set)))
def test_type_operation(expr, typename):
    sel = parse_selection(expr)
    assert isinstance(sel, typename)


@pytest.mark.parametrize('terms', (['foo', 'bar2'], ['a', 'b', 'c'],
                                   ['d', 'e', 'f', 'g']))
def test_and(terms):
    expr = ' and '.join(terms)
    assert parse_selection(expr) == terms


@pytest.mark.parametrize('terms', (('foo', 'bar2'), ('a', 'b', 'c'),
                                   ('d', 'e', 'f', 'g')))
def test_or(terms):
    expr = ' or '.join(terms)
    assert parse_selection(expr) == terms


@pytest.mark.parametrize('terms', (('foo', 'bar2'), ('a', 'b', 'c'),
                                   ('d', 'e', 'f', 'g')))
def test_not(terms):
    expr = 'not ({})'.format(' and '.join(terms))
    assert parse_selection(expr) == [{el} for el in terms]


def test_complex_expr():
    expr = "not food or (leaves and animals)"
    assert parse_selection(expr) == ({'food'}, ['leaves', 'animals'])


@pytest.mark.parametrize(
    'char',
    list(
        '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'*+,-./:;<=>?@[]^_`{|}~'
    ))
def test_possible_characters(char):
    sel = parse_selection("food and a{}".format(char))
    assert isinstance(sel, list)



"""Core classes and functions for handling state backends.

"""

import os
import sys
import fcntl
import warnings
import json
from functools import wraps
from contextlib import contextmanager
import six

if six.PY2:
    FileNotFoundError = IOError


class BaseFile(object):
    """Generic File object base class. Implements file locking and reloading
    methods.

    All files in datreant should be accessible by high-level methods
    without having to worry about simultaneous reading and writing by other
    processes. The File object includes methods for ensuring shared and
    exclusive locks are consistently applied before reads and writes,
    respectively. It handles any other low-level tasks for maintaining file
    integrity.

    Any child class will have to implement `_open_file_r` and `_open_file_w`.
    These function should return a file like object that has a `close` method.

    :Arguments:
        *filename*
            name of file on disk object corresponds to

    """

    def __init__(self, filename):
        self.filename = os.path.abspath(filename)
        self.handle = None
        self.fd = None
        self.fdlock = None

        # we apply locks to a proxy file to avoid creating an HDF5 file
        # without an exclusive lock on something; important for multiprocessing
        proxy = "." + os.path.basename(self.filename) + ".proxy"
        self.proxy = os.path.join(os.path.dirname(self.filename), proxy)

        # we create the file if it doesn't exist; if it does, an exception is
        # raised and we catch it; this is necessary to ensure the file exists
        # so we can use it for locks
        try:
            fd = os.open(self.proxy, os.O_CREAT | os.O_EXCL)
            os.close(fd)
        except OSError as e:
            # if we get the error precisely because the file exists, continue
            if e.errno == 17:
                pass
            else:
                raise

    def get_location(self):
        """Get File basedir.

        :Returns:
            *location*
                absolute path to File basedir

        """
        return os.path.dirname(self.filename)

    def _shlock(self, fd):
        """Get shared lock on file.

        Using fcntl.lockf, a shared lock on the file is obtained. If an
        exclusive lock is already held on the file by another process,
        then the method waits until it can obtain the lock.

        :Arguments:
            *fd*
                file descriptor
        """
        fcntl.lockf(fd, fcntl.LOCK_SH)

    def _exlock(self, fd):
        """Get exclusive lock on file.

        Using fcntl.lockf, an exclusive lock on the file is obtained. If a
        shared or exclusive lock is already held on the file by another
        process, then the method waits until it can obtain the lock.

        :Arguments:
            *fd*
                file descriptor
        """
        fcntl.lockf(fd, fcntl.LOCK_EX)

    def _unlock(self, fd):
        """Remove exclusive or shared lock on file.

        WARNING: It is very rare that this is necessary, since a file must be
        unlocked before it is closed. Furthermore, locks disappear when a file
        is closed anyway.  This method will remain here for now, but may be
        removed in the future if not needed (likely).

        :Arguments:
            *fd*
                file descriptor
        """
        fcntl.lockf(fd, fcntl.LOCK_UN)

    def _open_fd_r(self):
        """Open read-only file descriptor for application of advisory locks.

        Because we need an active file descriptor to apply advisory locks to a
        file, and because we need to do this before opening a file with
        the apprpriate interface (e.g. PyTables), we open
        a separate file descriptor to the same file and apply the locks
        to it.

        """
        self.fd = os.open(self.proxy, os.O_RDONLY)

    def _open_fd_rw(self):
        """Open read-write file descriptor for application of advisory locks.

        """
        self.fd = os.open(self.proxy, os.O_RDWR)

    def _close_fd(self):
        """Close file descriptor used for application of advisory locks.

        """
        # close file descriptor for locks
        os.close(self.fd)
        self.fd = None

    def _apply_shared_lock(self):
        """Apply shared lock.

        """
        self._open_fd_r()
        self._shlock(self.fd)
        self.fdlock = 'shared'

    def _apply_exclusive_lock(self):
        """Apply exclusive lock.

        """
        self._open_fd_rw()
        self._exlock(self.fd)
        self.fdlock = 'exclusive'

    def _release_lock(self):
        """Apply exclusive lock.

        """
        self._unlock(self.fd)
        self._close_fd()
        self.fdlock = None

    def _open_file_r(self):
        raise NotImplementedError

    def _open_file_w(self):
        raise NotImplementedError

    @contextmanager
    def read(self):
        # if we already have any lock, proceed
        if self.fdlock:
            yield self.handle
        else:
            self._apply_shared_lock()
            try:
                # open the file using the actual reader
                self.handle = self._open_file_r()
                yield self.handle
            finally:
                self.handle.close()
                self._release_lock()

    @contextmanager
    def write(self):
        # if we already have an exclusive lock, proceed
        if self.fdlock == 'exclusive':
            yield self.handle
        else:
            self._apply_exclusive_lock()

            # open the file using the actual writer
            self.handle = self._open_file_w()
            try:
                yield self.handle
            finally:
                self.handle.close()
                self._release_lock()

    def delete(self):
        """Delete this file and its proxy file.

        This file instance will be unusable after this operation.

        """
        with self.write():
            os.remove(self.filename)
            os.remove(self.proxy)


class FileSerial(BaseFile):
    """File object base class for serialization formats, such as JSON.

    """
    @property
    def _writebuffer(self):
        wbuffer = ".{}.buffer".format(os.path.basename(self.filename))
        return os.path.join(os.path.dirname(self.filename), wbuffer)

    def _open_file_r(self):
        return open(self.filename, 'r')

    def _open_file_w(self):
        return open(self._writebuffer, 'w')

    def read_file(self):
        """Return deserialized representation of file.

        """
        self._apply_shared_lock()

        self.handle = self._open_file_r()
        out = self._deserialize(self.handle)
        self.handle.close()

        self._release_lock()

        return out

    @contextmanager
    def read(self):
        # if we already have any lock, proceed
        if self.fdlock:
            yield self._state
        else:
            self._apply_shared_lock()
            try:
                self._pull_state()
                yield self._state
            except (FileNotFoundError, IOError):
                self._init_state(self)
                yield self._state
            finally:
                self._release_lock()

    @contextmanager
    def write(self):
        # if we already have an exclusive lock, proceed
        if self.fdlock == 'exclusive':
            yield self._state
        else:
            self._apply_exclusive_lock()
            try:
                self._pull_state()
            except IOError:
                self._init_state(self)
            try:
                yield self._state
                self._push_state()
            finally:
                self._release_lock()

    def _pull_state(self):
        self.handle = self._open_file_r()
        self._state = self._deserialize(self.handle)
        self.handle.close()

    def _deserialize(self, handle):
        """Deserialize full state from open file handle.

        Override with specific code for deserializing stream from *handle*.
        """
        raise NotImplementedError

    def _push_state(self):
        self.handle = self._open_file_w()
        self._serialize(self._state, self.handle)
        self.handle.close()
        os.rename(self._writebuffer, self.filename)

    def _serialize(self, state, handle):
        """Serialize full state to open file handle.

        Override with specific code for serializing *state* to *handle*.
        """
        raise NotImplementedError


class JSONFile(FileSerial):
    def __init__(self, filename, init_state, **kwargs):

        super(JSONFile, self).__init__(filename, **kwargs)

        # set _init_state function that given
        self._init_state = init_state

    def _deserialize(self, handle):
        return json.load(handle)

    def _serialize(self, state, handle):
        json.dump(state, handle)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['moldyn',
 'moldyn.data',
 'moldyn.processing',
 'moldyn.simulation',
 'moldyn.ui',
 'moldyn.ui.qt',
 'moldyn.utils',
 'moldyn.utils.datreant',
 'moldyn.utils.datreant.core',
 'moldyn.utils.datreant.scripts',
 'moldyn.utils.datreant.scripts.tests',
 'moldyn.utils.datreant.tests']

package_data = \
{'': ['*'], 'moldyn.simulation': ['templates/*']}

install_requires = \
['PyQt5>=5.11.3,<6.0.0',
 'asciitree>=0.3.3,<0.4.0',
 'fuzzywuzzy>=0.17.0,<0.18.0',
 'imageio_ffmpeg>=0.3.0,<0.4.0',
 'matplotlib>=3.1,<4.0',
 'moderngl>=5.5,<6.0',
 'numba>=0.53.1,<0.54.0',
 'numexpr>=2.6,<3.0',
 'numpy>=1.16,<2.0',
 'pathlib2>=2.3,<3.0',
 'pillow>=6.0,<7.0',
 'pyparsing>=2.4,<3.0',
 'pyqtgraph>=0.12.1,<0.13.0',
 'scandir>=1.10,<2.0',
 'scipy>=1.3,<2.0',
 'six>=1.12,<2.0']

entry_points = \
{'console_scripts': ['moldyn-gui = moldyn:gui']}

setup_kwargs = {
    'name': 'open-moldyn',
    'version': '0.1.0',
    'description': 'Tools for molecular dynamics simulation and analysis',
    'long_description': "# open-moldyn\n\nOpen Molecular Dynamics tools in Python 3\n\nread the doc on https://open-moldyn.readthedocs.io/en/latest/\n\n## Installation instructions\n\nThe package is on PyPI, to install use :\n```\npip install open-moldyn\n```\n\nTo update the package use :\n```\npip install -U open-moldyn\n```\n\nTo run open-moldyn's GUI use the script:\n```\nmoldyn-gui\n```\n\nIf it does not work in your terminal, \nmaybe your python installation is not in the PATH.\n\n",
    'author': 'Arthur Luciani, Alexandre Faye-Bedrin',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/open-moldyn/moldyn',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



# -*-encoding: utf-8 -*-
"""
Types personnalisés utilisés pour moldyn
"""

from struct import pack

class sList(list): # liste qui se convertir en bytes gentiment
    def __bytes__(self):
        return pack("".join((str(len(self)),"i")),*self)
    def __repr__(self):
        return "".join(("sList(", super().__repr__(), ")"))



# -*-encoding: utf-8 -*-
"""
Strain calculator.
Runs on CPU.
"""
import weakref
from pprint import pprint

from numpy.linalg import inv

"""
installer icc-rt et tbb sur les machines à processeur intel
"""

import numpy as np
import numba
import threading
import multiprocessing as mp
import ctypes



@numba.njit(nogil=True, cache=True)
def _iterate(current_pos, x, pos, posdt, NPART, rcut, LENGTH_X, LENGTH_Y, SHIFT_X, SHIFT_Y, X_PERIODIC, Y_PERIODIC):
    X = np.zeros((2,2))
    Y = np.zeros((2,2))
    eps = np.zeros((2,2))
    for n in range(NPART):
        if x==n:
            continue
        distxy = current_pos - pos[n, :]

        if X_PERIODIC:
            if distxy[0] < (-SHIFT_X):
                distxy[0] += LENGTH_X
            if distxy[0] > SHIFT_X:
                distxy[0] -= LENGTH_X

        if Y_PERIODIC:
            if distxy[1] < (-SHIFT_Y):
                distxy[1] += LENGTH_Y
            if distxy[1] > SHIFT_Y:
                distxy[1] -= LENGTH_Y

        if np.abs(distxy[0])<rcut and np.abs(distxy[1])<rcut:
            dist = np.sqrt(np.sum(distxy ** 2))
            if dist < rcut:
                distxydt = posdt[x] - posdt[n]
                if X_PERIODIC:
                    if distxydt[0] < (-SHIFT_X):
                        distxydt[0] += LENGTH_X
                    if distxydt[0] > SHIFT_X:
                        distxydt[0] -= LENGTH_X

                if Y_PERIODIC:
                    if distxydt[1] < (-SHIFT_Y):
                        distxydt[1] += LENGTH_Y
                    if distxydt[1] > SHIFT_Y:
                        distxydt[1] -= LENGTH_Y

                for i in range(2):
                    for j in range(2):
                        X[j,i] += distxy[i]*distxydt[j]
                        Y[j,i] += distxydt[i]*distxydt[j]

    eps[:] = X*inv(Y) - np.eye(2)
    return eps


def _par_iterate(current_pos, i, RCUT, NPART, LENGTH_X, LENGTH_Y, X_PERIODIC, Y_PERIODIC, SHIFT_X, SHIFT_Y):
    #ret = np.zeros((2,2))
    pos = np.array(_pos_array).reshape(NPART,2)
    posdt = np.array(_posdt_array).reshape(NPART,2)
    #ret += _iterate(current_pos, i, pos, posdt, NPART, RCUT, LENGTH_X, LENGTH_Y, SHIFT_X, SHIFT_Y, X_PERIODIC,
    #                Y_PERIODIC)
    return _iterate(current_pos, i, pos, posdt, NPART, RCUT, LENGTH_X, LENGTH_Y, SHIFT_X, SHIFT_Y, X_PERIODIC,
                    Y_PERIODIC)
    #return ret


def _p_compute_strain(pool, _pos, _posdt, eps_arr, RCUT, NPART, LENGTH_X, LENGTH_Y, X_PERIODIC, Y_PERIODIC, SHIFT_X, SHIFT_Y,
                      offset, end):

    pos = np.array(_pos).reshape(NPART,2)

    gen = ((pos[i,:], i, RCUT, NPART, LENGTH_X, LENGTH_Y, X_PERIODIC, Y_PERIODIC, SHIFT_X, SHIFT_Y)
           for i in range(offset, end))

    sortie = np.array(pool.starmap(_par_iterate, gen))

    eps_arr[:] = sortie[:]

_pos_array = None
_posdt_array = None


def initProcess(array1, array2):
    global _pos_array, _posdt_array
    _pos_array = array1
    _posdt_array = array2

class StrainComputeCPU:

    def __init__(self, consts, compute_npart=None, compute_offset=0):

        self.consts = dict()
        for key, item in consts.items():
            self.consts[key.upper()] = item

        self.compute_offset = max(0, compute_offset)

        self.npart = self.consts["NPART"]
        self.compute_npart = compute_npart or (self.consts["NPART"] - self.compute_offset)

        self.compute_npart = min(self.compute_npart, self.npart)

        self._EPS = np.zeros((self.npart, 2, 2), dtype=np.float32)

        self._thr_run = False

        self._POS = mp.Array(ctypes.c_float, self.npart * 2, lock=False)
        self._POSDT = mp.Array(ctypes.c_float, self.npart * 2, lock=False)
        self._pool = mp.Pool(mp.cpu_count(), initializer=initProcess, initargs=(self._POS, self._POSDT))

    def __del__(self):
        self._pool.close() # pour ne pas garder des processus ouverts pour rien

    def _compute_strain(self):
        RCUT = self.consts["RCUT"]
        NPART = self.consts["NPART"]
        LENGTH_X = self.consts["LENGTH_X"]
        LENGTH_Y = self.consts["LENGTH_Y"]
        X_PERIODIC = self.consts["X_PERIODIC"]
        Y_PERIODIC = self.consts["Y_PERIODIC"]

        SHIFT_X = LENGTH_X / 2
        SHIFT_Y = LENGTH_Y / 2
        end = NPART
        _p_compute_strain(self._pool, self._POS, self._POSDT, self._EPS, RCUT, NPART, LENGTH_X, LENGTH_Y, X_PERIODIC,
                          Y_PERIODIC, SHIFT_X, SHIFT_Y, self.compute_offset, end)

    def _join_thr(self):
        if self._thr_run:
            self._thread.join()
            self._thr_run = False

    def set_post(self, pos):
        self._POS[:] = pos.reshape((self.npart * 2,))

    def set_posdt(self, pos):
        self._POSDT[:] = pos.reshape((self.npart * 2,))

    def get_eps(self):
        self._join_thr()
        return self._EPS[:]

    def compute(self):
        self._thr_run = True
        self._thread = threading.Thread(target=self._compute_strain)
        self._thread.start()